import unittest
from warnings import warn
import pyrosetta
pyrosetta.init(extra_options='-mute all')  # required for test
from rdkit_to_params import Params
from rdkit import Chem

class MyTestCase(unittest.TestCase):
    def test_load(self):
        p = Params.load('example/official_PHE.params')
        self.assertEqual(p.NAME, 'PHE')

    def test_round(self):
        p = Params.load('example/official_PHE.params')
        p.IO_STRING[0].name3 = 'PHX'
        p.IO_STRING[0].name1 = 'Z'
        p.AA = 'UNK'  # If it's not one of the twenty (plus extras), UNK!
        del p.ROTAMER_AA[0]
        p.rename_atom(' CB ', ' CX ')  # this renames
        p.dump('fake.params')
        pose = p.test()
        buffer = pyrosetta.rosetta.std.stringbuf()
        pose.dump_pdb(pyrosetta.rosetta.std.ostream(buffer))
        pdbblock = buffer.str()
        self.assertIsNotNone(Chem.MolFromPDBBlock(pdbblock))

    def test_smiles(self):
        p = Params.from_smiles('*C(=O)C(Cc1ccccc1)[NH]*',  # recognised as amino acid.
                               name='PHX',  # optional.
                               atomnames={3: 'CZ'}  # optional, rando atom name
                               )
        self.assertTrue(p.is_aminoacid())

    def test_renames(self):
        p = Params.from_smiles('CC(ONC)O', atomnames=['CA', 'CB', 'OX', 'ON', 'CX', 'CG'])
        p.rename_atom_by_name('CA', 'CZ')
        self.assertEqual(p.mol.GetAtomWithIdx(0).GetPDBResidueInfo().GetName(), ' CZ ')
        self.assertEqual(p.ATOM[0].name, ' CZ ')
        p.test()



if __name__ == '__main__':
    unittest.main()



########################################################################################################################
__doc__ = \
    """
The main class here is ``_ParamsInitMixin``, which simply adds the __init__ method...
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################


from .entries import Entries

class _ParamsInitMixin:

    # This is solely for my sanity/notebook and does not and should not be used.
    ordering = ['NAME', 'IO_STRING', 'TYPE', 'AA', 'ROTAMER_AA', '#' 'ATOM', 'ATOM_ALIAS', 'BOND', 'CUT_BOND',
                'CHI', 'CONNECT', 'ADD_RING', 'PROPERTIES', 'METAL_BINDING_ATOMS', 'FIRST_SIDECHAIN_ATOM',
                'RAMA_PREPRO_FILENAME', 'ACT_COORD_ATOMS',
                'NBR_ATOM', 'NBR_RADIUS', 'ICOOR_INTERNAL', 'PDB_ROTAMERS']

    def __init__(self):
        self.AA = Entries.from_name('AA')  # for ligands
        self.TYPE = Entries.from_name('TYPE') #POLYMER or LIGAND
        self.ROTAMER_AA = Entries.from_name('ROTAMER_AA')
        self.IO_STRING = Entries.from_name('IO_STRING')
        self.comments = Entries.from_name('#')
        self.ATOM = Entries.from_name('ATOM')
        self.ATOM_ALIAS = Entries.from_name('ATOM_ALIAS')
        self.BOND = Entries.from_name('BOND')
        self.CUT_BOND = Entries.from_name('CUT_BOND')
        self.CONNECT = Entries.from_name('CONNECT')
        self.CHI = Entries.from_name('CHI')
        self.NBR_ATOM = Entries.from_name('NBR_ATOM')
        self.NBR_RADIUS = Entries.from_name('NBR_RADIUS')
        self.ICOOR_INTERNAL = Entries.from_name('ICOOR_INTERNAL')
        self.ADD_RING = Entries.from_name('ADD_RING')
        self.PROPERTIES = Entries.from_name('PROPERTIES')
        self.FIRST_SIDECHAIN_ATOM = Entries.from_name('FIRST_SIDECHAIN_ATOM')
        self.RAMA_PREPRO_FILENAME = Entries.from_name('RAMA_PREPRO_FILENAME')
        self.METAL_BINDING_ATOMS = Entries.from_name('METAL_BINDING_ATOMS')
        self.ACT_COORD_ATOMS = Entries.from_name('ACT_COORD_ATOMS')
        self.PDB_ROTAMERS = Entries.from_name('PDB_ROTAMERS') #TODO it would be nice to check this file exists.
        ## RDKit route specific
        self.mol = None
        self.generic = False
        self._rtype = []

    @property
    def fields(self):
        """
        This operates under the assumption that the user may have added extra uppercase entries!

        :return:
        """
        # NAME is a dynamic property.
        return ['NAME'] + [k for k in self.__dict__ if k.upper() == k]


from __future__ import annotations

########################################################################################################################
__doc__ = \
    """
This contains ``make_constraint`` which is creates a constraint file.
It is completely independent and different in style because it was different.
It is not integral to the conversion, it's just a utility.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

from typing import Tuple, List, Union, Optional, Dict
from rdkit import Chem
from rdkit.Chem import AllChem

import json


class Constraints:
    def __init__(self, smiles: Tuple[str, str],
                 names: List[str],
                 ligand_res: Union[str, int],
                 target_res: Union[str, int]):
        """
        Generate the required constraints for a covalent bond as CONN1 does not deal with distance and torsions.

        Give smiles of the two sides (with ``*``) and the names, residue numbers convert.
        It requires 2 atoms on each side in addition to the attachment point.
        Note that the atom names are stored in a non-standard way out of laziness. in the Atom prop '_AtomName'.
        The instance has the following attributes

        * ``smiles``: stored input smiles string
        * ``names``: stored input list of names
        * ``ligand_res``: stored input ligand residue
        * ``target_res``: stored input protein residue
        * ``cov_template``: Chem.Mol from first smiles.
        * ``target_template``: Chem.Mol from first smiles
        * ``combo``: combined templates
        * ``atom_pair_constraint``: AtomPair
        * ``angle_constraint``: Angle . NB. this is two lines.
        * ``dihedral_constraint``: dihedral
        * ``coordinate_constraint``: coordinate (see make_coordinate_constraint methods)
        * ``custom_constaint``: user added these

        Class methods:
        * ``assign_names(mol, list)`` classmethods that assigns names (dodgy '_AtomName') to a mol in place.
        * ``nominalise(mol)`` propagate the nonstandard names (good for saving sdf)
        * ``join_by_dummy(molA, molB)`` classmethods that returns a joined molecule



        >>> Constraints(smiles=('*C(=N)', '*SC'), names= ['*', 'CX', 'NY', '*', 'SG', 'CB'], ligand_res= '1B', target_res='145A')

        AtomPair SG 145A CX 1B HARMONIC 1.74 0.2

        Angle CB 145A SG 145A CX 1B HARMONIC 1.73 0.35

        Angle SG 145A CX 1B NY 1B HARMONIC 2.08 0.35

        Dihedral CB 145A SG 145A CX 1B NY 1B CIRCULARHARMONIC -3.14 0.35

        :param smiles: a tuple/list of two string. The first is the ligand, the second is the peptide.
        :param names: a list of atom names. The '*' will need a name -but will be ignored-, but not the H.
        :param ligand_res: ligand residue in pose or PDB format (12 vs. 12A)
        :param target_res:  peptide residue in pose or PDB format (12 vs. 12A)
        """
        # store inputs
        self.smiles = smiles
        self.names = names
        self.ligand_res = ligand_res
        self.target_res = target_res
        # Convert the smiles
        self.cov_template = Chem.MolFromSmiles(smiles[0])
        self.target_template = Chem.MolFromSmiles(smiles[1])
        # check all is good
        assert self.cov_template.GetNumAtoms() + self.target_template.GetNumAtoms() == len(names), \
            f'{len(names)} were provided but the SMILES have {self.cov_template.GetNumAtoms()}+{self.target_template.GetNumAtoms()} atoms.'
        # assign names
        self.assign_names(self.cov_template, self.names[:self.cov_template.GetNumAtoms()])
        self.assign_names(self.target_template, self.names[self.cov_template.GetNumAtoms():])
        self.cov_con_name = self.get_conn(self.cov_template).GetProp('_AtomName')
        self.target_con_name = self.get_conn(self.target_template).GetProp('_AtomName')
        # combine and embed
        self.combo = self.join_by_dummy(self.cov_template, self.target_template)
        conf = self._get_conformer(self.combo)
        # get key atoms
        cov_con = self.get_atom(self.combo, self.cov_con_name)
        cov_fore = cov_con.GetNeighbors()[0]
        cov_fore_name = cov_fore.GetProp('_AtomName')
        target_con = self.get_atom(self.combo, self.target_con_name)
        target_fore = target_con.GetNeighbors()[0]
        target_fore_name = target_fore.GetProp('_AtomName')
        # do maths
        ## Note: constraint is in Radian not Degree...
        dist = Chem.rdMolTransforms.GetBondLength(conf, cov_con.GetIdx(), target_con.GetIdx())
        angle_target = Chem.rdMolTransforms.GetAngleRad(conf, target_fore.GetIdx(), target_con.GetIdx(),
                                                        cov_con.GetIdx())
        angle_covalent = Chem.rdMolTransforms.GetAngleRad(conf, target_con.GetIdx(), cov_con.GetIdx(),
                                                          cov_fore.GetIdx())
        dihedral = Chem.rdMolTransforms.GetDihedralRad(conf, target_fore.GetIdx(), target_con.GetIdx(),
                                                       cov_con.GetIdx(),
                                                       cov_fore.GetIdx())
        self.atom_pair_constraint = f'AtomPair {self.target_con_name} {target_res} {self.cov_con_name} {ligand_res} '+\
                                    f'HARMONIC {dist:.2f} 0.2\n'
        self.angle_constraint = f'Angle {target_fore_name} {target_res} {self.target_con_name} {target_res} '+\
                                f'{self.cov_con_name} {ligand_res} HARMONIC {angle_target:.2f} 0.35\n' + \
                                f'Angle {self.target_con_name} {target_res} {self.cov_con_name} {ligand_res} '+\
                                f' {cov_fore_name} {ligand_res} HARMONIC {angle_covalent:.2f} 0.35\n'
        self.dihedral_constraint = f'Dihedral {target_fore_name} {target_res} {self.target_con_name} {target_res} '+\
                                f'{self.cov_con_name} {ligand_res} {cov_fore_name} {ligand_res} '+\
                                f'CIRCULARHARMONIC {dihedral:.2f} 0.35\n'
        self.coordinate_constraint = ''
        self.custom_constraint = ''

    # =================== dependent methods ============================================================================

    def _get_conformer(self, mol):
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol)
        AllChem.MMFFOptimizeMolecule(mol)
        return mol.GetConformer()

    def __str__(self):
        # make
        constraints = [self.atom_pair_constraint,
                       self.angle_constraint,
                       self.dihedral_constraint,
                       self.custom_constraint]
        return ''.join(constraints)

    def dumps(self):
        return str(self)

    def dump(self, filename):
        with open(filename, 'w') as w:
            w.write(str(self))

    @classmethod
    def _get_new_index(self, mol: Chem.Mol, index: int) -> int:
        for atom in mol.GetAtoms():
            if atom.GetDoubleProp('_originalIdx') == index:
                return atom.GetIdx()

    def get_atom(self, mol: Chem.Mol, name: str) -> Chem.Atom:
        for atom in mol.GetAtoms():
            if atom.HasProp('_AtomName') and name == atom.GetProp('_AtomName'):  # Nonstandard. do not copy
                return atom
        else:
            raise ValueError(f'Atom {name} not found')

    # =================== utilities ====================================================================================

    @classmethod
    def mock(cls):
        """
        This is an empty instance.

        :return:
        """
        self = cls.__new__(cls)
        self.atom_pair_constraint = ''
        self.angle_constraint = ''
        self.dihedral_constraint = ''
        self.coordinate_constraint = ''
        self.custom_constraint = ''
        return self

    @classmethod
    def assign_names(cls, mol: Chem.Mol, names: List[str]) -> None:
        """
        Stores names of atoms as given in the list.
        totally non-standard way. PDBInfo is correct way. But too much effort.
        """
        for name, atom in zip(names, mol.GetAtoms()):
            atom.SetProp('_AtomName', name)  # Nonstandard. do not copy

    @classmethod
    def nominalise(cls, mol):
        """
        If the mol has PDBResidueInfo it will fill the nonstandard mol prop _AtomNames and the atom prop _AtomName.
        If it has mol prop _AtomNames will propagate them down to the atoms and viceversa. But will not fill
        ``PDBResidueInfo`` because there is too much metadata required.

        :param mol:
        :return:
        """
        if mol.GetAtomWithIdx(0).GetPDBResidueInfo() is not None: #has PDB info!
            names = []
            for atom in mol.GetAtoms():
                info = atom.GetPDBResidueInfo()
                if info is not None:
                    n = info.GetName()
                    atom.SetProp('_AtomName', n)
                    names.append(n)
                else:
                    names.append(None)
            mol.SetProp('_AtomNames', json.dumps(names))
        elif mol.HasProp('_AtomNames'):
            names = json.loads(mol.GetProp('_AtomNames'))
            for name, atom in zip(mol.GetAtoms(), names):
                if name is not None:
                    atom.SetProp('_AtomName', name)
        else:
            names = [atom.GetProp('_AtomName') if atom.HasProp('_AtomName') else None for atom in mol.GetAtoms()]
            if len([n is not None for n in names]) == 0:
                raise ValueError('No type of Atom naming. Use `Params.load_mol`  to fix this!')
            mol.SetProp('_AtomNames', json.dumps(names))

    @classmethod
    def get_conn(cls, mol: Chem.Mol) -> Chem.Atom:
        """
        Get connecting atom of mol.
        """
        for atom in mol.GetAtoms():
            if atom.GetSymbol() == '*':
                return atom.GetNeighbors()[0]
        else:
            raise ValueError('Dummy atom not found')

    @classmethod
    def join_by_dummy(cls, a: Chem.Mol, b: Chem.Mol) -> Chem.Mol:
        # So I was worried that joining by the connect neighbour and then deleting the dummy
        # may cause issues of valence. So I did it this more convoluted way.
        # but did not check first... and this approach leads to sanitisation...
        for atom in a.GetAtoms():
            atom.SetDoubleProp('_originalIdx', atom.GetIdx())
        conn = cls.get_conn(a)
        mod_a = Chem.DeleteSubstructs(a, Chem.MolFromSmiles('*'))
        i = cls._get_new_index(mod_a, conn.GetIdx())
        combo = Chem.ReplaceSubstructs(b, Chem.MolFromSmiles('*'), mod_a, replacementConnectionPoint=i)[0]
        combo.UpdatePropertyCache()
        Chem.SanitizeMol(combo)
        return combo

    @classmethod  # /bound method!
    def make_coordinate_constraints(celf, mol, ligand_res, ref_res, ref_atomname='CA',
                                    stdevs: Optional[Dict[Union[str,int], float]] = None) -> Constraints:
        """
        Returns an instance (or the same instance if called as a bound method) with added ``.coordinate_constraint``.
        If no stdevs are supplied all non dummy atoms have a HARMONIC with std 1.

        :param mol:
        :param ligand_res:
        :param ref_res:
        :param ref_atomname:
        :param stdevs: stdev to use for HARMONIC. dictionary with keys either integer (AtomIdx) or str (atomname)
        :type stdevs: Union[str,int], float]]
        :return:
        """
        if hasattr(celf, '__class__'):
            self = celf
        else:
            cls = celf
            self = cls.mock()
        self.nominalise(mol)
        lines = []
        conf = mol.GetConformer()
        # issue of 4 char padded names.
        sstdevs = {k.strip if isinstance(k, str) else k: v for k, v in stdevs.items()}
        for i, atom in enumerate(mol.GetAtoms()):
            if atom.GetSymbol() == '*':
                continue
            elif sstdevs is not None:
                n = atom.GetProp('_AtomName').strip() if atom.HasProp('_AtomName') else None
                if i in sstdevs and sstdevs[i] not in (float('inf'), float('nan'), None):
                    w = sstdevs[i]
                elif n in sstdevs and sstdevs[n] not in (float('inf'), float('nan'), None):
                    w = sstdevs[n]
                else:
                    continue
            else:
                w = 1
            pos = conf.GetAtomPosition(i)
            lines.append(f'CoordinateConstraint {atom.GetPDBResidueInfo().GetName()} {ligand_res} ' + \
                         f'{ref_atomname} {ref_res} ' + \
                         f'{pos.x} {pos.y} {pos.z} HARMONIC 0 {w}\n')
        self.coordinate_constraint += ''.join(lines)
        return self

    @classmethod  # /bound method!
    def make_inverse_coordinate_constraints_by_neighbours(cls, mol,
                                                          ligand_res,
                                                          unfixed: List[str],
                                                          ref_res,
                                                          ref_atomname='CA',
                                    ) -> Constraints:
        """
        Given a list of indices constraint their distant neighbours.
        This is basically for modify those atoms. E.g. given one structure, make a variant.

        :param mol:
        :param ligand_res:
        :param unfixed:
        :param ref_res:
        :param ref_atomname:
        :return:
        """
        def fill(atom, n):
            for natom in atom.GetNeighbors():
                if natom.GetIdx() in stdevs:
                    return None
                elif natom.GetSymbol() in ('*', 'H'):
                    return None
                else:
                    stdevs[natom.GetIdx()] = n
                    fill(natom, n/2)
                    return None

        cls.nominalise(mol)
        stdevs = {}
        unfixed = [u.strip() for u in unfixed]
        for unname in unfixed:
            a = [atom for atom in mol.GetAtoms() if atom.GetProp('_AtomName').strip() == unname.strip()]
            if len(a) == 0:
                raise ValueError(f'{unname} does not exist')
            elif len(a) > 1:
                raise ValueError(f'{unname} is present more than once??!')
            else:
                unatom = a[0]
                stdevs[unatom.GetIdx()] = float('inf')
                fill(unatom, 4)
        return cls.make_coordinate_constraints(mol, ligand_res, ref_res, ref_atomname, stdevs)


if __name__ == '__main__':
    c = Constraints(smiles=('*C(=N)', '*SC'),
                    names=['*', 'CX', 'NY', '*', 'SG', 'CB'],
                    ligand_res='1B',
                    target_res='145A')
    print(c)



from rdkit import Chem
from rdkit.Chem import rdFMCS, AllChem

from typing import Sequence, List, Dict, Optional, Any, Union
from warnings import warn
import re, logging


########################################################################################################################
__doc__ = \
    """
The main class here is ``_RDKitRenameMixin``, which adds the various atom renaming methods.

    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

class _RDKitRenameMixin:
    log = logging.getLogger(__name__)

    # ============= overridden =========================================================================================

    def __init__(self):
        # will be overridden. Just here for typehinting...
        self.log.critical('WRONG INIT CALLED')
        self.mol = Chem.Mol()
        self.NAME = ''
        self.is_aminoacid = lambda: True # main class
        self.polish = lambda: None # covert class. makes sure all atoms have the same rdkit residue info
        self.rename_atom = lambda oldname, newname, overwrite=False: None
        self.pad_name = lambda name: name

    # ============= set and get ========================================================================================

    def _get_PDBInfo_atomname(self, atom, throw=True) -> str:
        info = atom.GetPDBResidueInfo()
        if info is not None:
            return info.GetName()
        elif throw:
            raise ValueError('Atoms changed but `fix_mol` was not called.')
        else:
            return ''

    def _set_PDBInfo_atomname(self, atom, name, overwrite=False):
        info = atom.GetPDBResidueInfo()
        if info is None:
            isHeteroAtom = not self.is_aminoacid()
            atom.SetMonomerInfo(Chem.AtomPDBResidueInfo(atomName=self.pad_name(name),
                                                        serialNumber=atom.GetIdx(),
                                                        residueName=self.NAME,
                                                        isHeteroAtom=isHeteroAtom))
            return name
        elif info.GetName() == name:
            return name
        elif overwrite:
            info.SetName(name)
            return name
        else:
            return info.GetName()


    def get_atom_by_name(self, name):
        for atom in self.mol.GetAtoms():
            info = atom.GetPDBResidueInfo()
            if info and info.GetName().strip() == name.strip():
                return atom
        else:
            raise ValueError(f'Name {name} not found.')

    # ============= retype method ======================================================================================

    def retype_by_name(self, mapping: Dict[str, str]):
        """
        Renames the atom names in ``self.mol`` by changing key to value of the mapping dictionary.

        :param mapping: old name to new name
        :return:
        """
        for name, rtype in mapping.items():
            if rtype is None:
                continue
            atom = self.get_atom_by_name(name)
            if atom is not None:
                atom.SetProp('_rType', rtype)
            else:
                self.log.warning(f'Could not find "{name}"')
        for entry in self.ATOM:
            if entry.name.strip() == name.strip():
                entry.rtype = rtype
                break
        else:
            pass

    # ============= rename methods =====================================================================================

    def rename(self, atomnames: Union[None, dict, list, str, Chem.Mol]) -> None:
        """
        Rename options for atomnames:

        * None: nothing
        * Dict: ``rename_from_dict``
        * List: ``rename_from_list``
        * str in '1:XX, 3:YY' format: ``rename_from_dict`` via ``rename_from_str``
        * str in 'XX,YY' format: ``rename_from_list`` via ``rename_from_str``
        * Chem.Mol: ``rename_from_template``

        :param atomnames: renaming values
        :return:
        """
        if atomnames is None:
            pass
        elif isinstance(atomnames, dict):
            self.rename_from_dict(atomnames)
        elif isinstance(atomnames, list):
            self.rename_from_list(atomnames)
        elif isinstance(atomnames, str):
            self.rename_from_str(atomnames)
        elif isinstance(atomnames, Chem.Mol):
            self.rename_from_template(atomnames)
        else:
            raise ValueError(f'atomnames is not None, dict, list, str or Chem.Mol but {type(atomnames)}')


    def rename_by_substructure(self, substructure: Chem.Mol, atomnames: Sequence[str]) -> List[str]:
        """
        Assigns to the atoms in self.mol the names based on the backbone template and the names variable.
        See ``_fix_atom_names`` for example usage.
        Changes also params.

        :param substructure:
        :param atomnames: the list oof new names. Falsey names will not be set.
        :return: the list of the old names.. why? May change in future.
        """
        assert self.mol.HasSubstructMatch(substructure), 'Bad backbone match'
        originals = []
        substructure = Chem.AddHs(substructure, explicitOnly=True)
        for name, idx in zip(atomnames, self.mol.GetSubstructMatch(substructure)):
            atom = self.mol.GetAtomWithIdx(idx)
            oldname = self._get_PDBInfo_atomname(atom, throw=False)
            self.rename_atom(atom, name)  # alters entry
            originals.append(oldname)
        return originals

    def rename_from_str(self, atomnames:str):
        """
        * '1:XX, 3:YY' format: calls ``rename_from_dict``
        * 'XX,YY' format: calls ``rename_from_list``

        :param atomnames: str
        :return:
        """
        if ':' in atomnames:
            self.rename_from_dict({int(k): v for k, v in re.findall(r'(\d+):([\s\w]{1,4})', atomnames)})
        else:
            self.rename_from_list(atomnames.split(','))

    def rename_from_template(self, template: Chem.Mol, overwrite:bool=True):
        """
        Assigns to the atoms in self.mol the names based on the template, which does not need to be a perfect match.
        See ``_fix_atom_names`` for example usage.
        Does not change the Params.

        :param template: mol object with atom names

        :return: None for now.
        """
        AllChem.SanitizeMol(template) #this is where half my issues come from.
        mcs = rdFMCS.FindMCS([self.mol, template],
                             atomCompare=rdFMCS.AtomCompare.CompareElements,
                             bondCompare=rdFMCS.BondCompare.CompareAny,
                             ringMatchesRingOnly=True)
        common = Chem.MolFromSmarts(mcs.smartsString)
        for acceptor, donor in zip(self.mol.GetSubstructMatch(common), template.GetSubstructMatch(common)):
            a_atom = self.mol.GetAtomWithIdx(acceptor)
            d_atom = template.GetAtomWithIdx(donor)
            info = d_atom.GetPDBResidueInfo()
            if info:
                self.rename_atom(a_atom, info.GetName(), overwrite=overwrite)
            else:
                self.log.debug.info(f'No info in template for atom {d_atom.GetSymbol()} #{donor}')

    def rename_from_dict(self, atomnames: Dict[int, str]):
        """
        Renames the ``self.mol`` atom names by a dict that has key atom idx and value name

        :param atomnames: idx -> new name
        :return:
        """
        # sanity check
        assert len(set(atomnames.values())) == len(atomnames), 'Atom Names are repeated.'
        if self.mol.GetNumAtoms() > len(atomnames):
            self.log.info('There are more atoms in mol than were provided.')
        elif self.mol.GetNumAtoms() < len(atomnames):
            #raise ValueError('There are less atoms in mol than were provided.')
            pass # this is fine.
        # operate
        for k, v in atomnames.items():
            self._set_PDBInfo_atomname(self.mol.GetAtomWithIdx(k), v, overwrite=True)
        self.polish_mol()

    def rename_from_list(self, atomnames: List[str]):
        """
        Renames the ``self.mol`` atom names based on the order in list. If None it is skipped.

        :param atomnames: list of unique names.
        :return:
        """
        # sanity check
        assert len(set(atomnames)) == len(atomnames), 'Atom Names are repeated.'
        if self.mol.GetNumAtoms() > len(atomnames):
            self.log.info('There are more atoms in mol than were provided.')
        elif self.mol.GetNumAtoms() < len(atomnames):
            raise ValueError('There are less atoms in mol than were provided.')
        # rename
        r = {i: name for i, name in enumerate(atomnames) if name is not None}
        self.rename_from_dict(r)

    # ============= class method =======================================================================================

    @classmethod
    def add_names(cls, mol: Chem.Mol, atomnames: Union[None, dict, list, str, Chem.Mol], name: Optional[str] = None) -> Chem.Mol:
        """
        Quick way to add atom names to a mol object --adds them the normal way.

        :param mol: Chem.Mol, will actually be edited in place.
        :param names: list/dict/str/chem of unique names.
        :param name: 3letter code for the molecule.
        :return: the mol
        """
        self = cls()

        self.log.debug(f'`add_names` called...')
        if name is not None:
            self.NAME = name
        self.mol = mol
        self.TYPE.append('LIGAND')
        self.fix_mol()
        self.rename(atomnames)
        return self.mol


########################################################################################################################
__doc__ = \
    """
    The main class here is ``Params``. All underscore base classes are not meant to be used standalone.
    ``Entries`` is the class for an list of entries of the same kind.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################


from warnings import warn
import os, re, logging
from typing import Union

#################### base classes ######################################################################################

from ._io_mixin import _ParamsIoMixin # in turn inherits _ParamsInitMixin
from .entries import Entries

#################### pyrosetta #########################################################################################
try:
    from ._pyrosetta_mixin import _PoserMixin, pyrosetta
except ImportError:
    warn('PyRosetta is required for the ``test`` method', category=ImportWarning)


    class _PoserMixin:
        pass

#################### rdkit #############################################################################################
try:
    from .rdkitside import _RDKitMixin
    from .constraint import Constraints
    from rdkit import Chem
except ImportError:
    warn('RDkit is required for ``from_mol`` stuff and ``Constraints``', category=ImportWarning)

    class Chem:
        Atom = None

    class _RDKitMixin:
        pass


#################### main class ########################################################################################


class Params(_ParamsIoMixin, _RDKitMixin, _PoserMixin):
    """

    ``Params`` creates and manipulates params files. It can handles several types of params operations,
    such as "atom name surgery" and ``rdkit.Chem.Mol`` to a params file.

    ## Key methods

    * ``Params.load(filename)`` will instantiate from file.
    * ``Params.from_mol(mol)`` will instantiate from ``Chem.Mol``
    * ``p.dump(filename)`` will save a file.
    * ``loads`` and ``dumps``for strings.
    * ``p.fields`` will return all header fields.
    * ``p.test`` tests the params file in PyRosetta.
    * ``p.rename_atom(old, new)`` changes an atom name

    ## Attributes

    The attributes are generally the uppercase line headers, with a few exceptions.

    * `.comments` is for # lines
    * "BOND_TYPE" and "BOND" are merged into ``.BOND``.
    * "UPPER", "LOWER" and "CONNECT" are merged into ``.CONNECT``

    With the exception of ``.NAME`` which depends on ``.IO_STRING`` basically
    all the header type attributes are actually instances of the class `Entries`, which holds a sequence of specific entries.
    see `entry.py` for the properties of each.
    These can be a singleton, such as `.IO_STRING` which when a new line is added it gets overwritten instead, or not like say `.ATOM`.
    That is to say that ``.ATOM[0]`` will give the first atom as expected, but this has to be done for ``.IO_STRING[0]`` too.

    Atomnames...

    * ``p.get_correct_atomname`` will return the 4 letter name of the atom with nice spacing.
    * ``p.rename_atom`` will change one atomname to a new one across all entries.
    * ``BOND``, ``CHI``, ``CUT_BOND`` entries store 4 char atomnames as ``.first``, ``.second``, ``.third``, ``.fourth``.
    * ``ICOOR_INTERNAL`` entries store 5 char atomnames as ``.child``,``.parent``,``.second_parent``,``.third_parent``.
    * ``ATOM_ALIAS``, ``NBR_ATOM``, ``FIRST_SIDECHAIN_ATOM``, ``ADD_RING`` are just ``entries.GenericEntries`` instances, where ``.body`` is a string which will contain the atomname.
    * ``METAL_BINDING_ATOMS``, ``ACT_COORD_ATOMS`` are ``entries.GenericListEntries`` instances where ``.values`` is a list of string with maybe atomnames.

    ## Inheritance

    It inherits several class,
    which are not not mean to be used standalone, except for testing.

    The pyrosetta and rdkit functionality are dependent on these being installed.

    * ``_ParamsIoMixin`` adds read write, and inherits
    * ``_ParamsInitMixin`` which adds the basics.
    * ``_PoserMixin`` is a base that adds pyrosetta functionality if avaliable.
    * ``_RDKitCovertMixin``, which adds rdkit ``from_mol`` conversion functionality, the class is split in two, the other part being
    * ``_RDKitParamsPrepMixin``, which prepares the molecule for `_RDKitCovertMixin.from_mol``.

    """
    log = logging.getLogger(__name__)

    @property
    def NAME(self):
        if len(self.IO_STRING):
            return self.IO_STRING[0].name3
        else:
            self.log.warning('Attempted access to empty IO_STRING/NAME')
            return 'XXX'

    @NAME.setter
    def NAME(self, name):
        if len(self.IO_STRING) == 0:
            self.IO_STRING.append(f'{name} Z')
        else:
            self.IO_STRING[0].name3 = name

    def is_aminoacid(self):
        if len(self.TYPE) == 0:
            self.TYPE.append('LIGAND')
        return self.TYPE[0].body == 'POLYMER'

    def validate(self):
        self.log.critical('This is not finished.')
        if 'CANONICAL_AA' in self.PROPERTIES[0].values:
            assert self.AA != 'UNK', 'CANONICAL_AA property requires a AA type not UNK'
        if 'METALBINDING' in self.PROPERTIES[0].values:
            assert len(self.METAL_BINDING_ATOMS) > 0, 'METALBINDING property requires METAL_BINDING_ATOMS'
        assert os.path.exists(self.PDB_ROTAMERS.strip()), f'PDB_ROTAMERS file {self.PDB_ROTAMERS} does not exist'
        # etc.

    def get_correct_atomname(self, name: str) -> str:
        """
        Given a name, gets the correctly spaced out one as appears in the ATOM entry.
        To pad out a name use pad_name
        This has nothing to do with ``._get_PDBInfo_atomname`` which just returns the atom name from a ``Chem.Atom``.

        :param name: dirty name
        :return: correct name
        """
        name = name.upper()
        # find in atom.
        for atom in self.ATOM:
            if atom.name == name:
                return name
        else:
            for atom in self.ATOM:
                if atom.name.strip() == name.strip():
                    return atom.name
            else:
                raise ValueError(f'{name} is not a valid atom name (does not appear in the entries)')

    def rename_atom(self, atom_or_atomname: Union[str, 'Chem.Atom'], newname: str, overwrite=True) -> Union[str, None]:
        """
        rename an atom by atomname or Chem.Atom (the former just calls ``rename_atom_by_name`` as is just for legacy)

        :param atom_or_atomname:
        :param newname:
        :return:
        """
        # sanity
        if newname is None:
            return None
        try:
            if self.mol:
                atom = self.get_atom_by_name(newname)
                if isinstance(atom_or_atomname, str):
                    raise AssertionError(f'New name {newname} already exists')
                elif isinstance(atom_or_atomname, Chem.Atom) and atom_or_atomname.GetIdx() != atom.GetIdx():
                    raise AssertionError(f'New name {newname} already exists')
                else:
                    pass # already changed.
            if len(self.ATOM) > 0:
                self.get_correct_atomname(newname)
        except ValueError:
            pass # absent
        # change.
        if isinstance(atom_or_atomname, str): #atom name
            oldname = atom_or_atomname
            return self.rename_atom_by_name(oldname, newname)
        elif isinstance(atom_or_atomname, Chem.Atom):
            atom = atom_or_atomname
            oldname = self._get_PDBInfo_atomname(atom, throw=False)
            if oldname:
                return self.rename_atom_by_name(oldname, newname)  # alters entry & rdkit
            else:
                return self._set_PDBInfo_atomname(atom, newname, overwrite=overwrite) # alters rdkit
        else:
            raise TypeError(f'{type(atom_or_atomname)} is not a string or atom')


    def rename_atom_by_name(self, oldname: str, newname: str) -> str:
        """
        Change the atom name from ``oldname`` to ``newname`` and returns the 4 char ``newname``.

        :param oldname: atom name, preferably 4 char long.
        :param newname: atom name, preferably 4 char long.
        :return: 4 char newname
        """
        if newname is None:
            return None
        elif oldname == newname:
            return newname
        else:
            # rdkit mol
            if self.mol:
                atom = self.get_atom_by_name(oldname)
                newname = self.pad_name(newname, atom)
                atom.GetPDBResidueInfo().SetName(newname)
            else:
                newname = self.pad_name(newname)
            # params
            self._rename_atom_in_entries(oldname, newname)
            return newname

    def _rename_atom_in_entries(self, oldname, newname):
        # if params is not filled nothing happens.
        # check if it is a connect atom
        if len(self.ATOM) == 0:
            return None # N/A: unparameterise atm.
        elif oldname.strip() == 'CONN':
            pass  # ...
        elif oldname.strip() in ('CONN1', 'CONN2', 'CONN3', 'LOWER', 'UPPER'):
            for conn in self.CONNECT:
                if conn.connect_name.strip() == oldname.strip():
                    conn.connect_name = newname
                    # TODO fix this properly. LOWER UPPER CONN3 should be the preferred order.
                    break
            else:
                raise ValueError(f'{oldname} does not appear amid the connections')
            for entry in self.ICOOR_INTERNAL:
                for key in 'child', 'parent', 'second_parent', 'third_parent':
                    if getattr(entry, key) == oldname.rjust(5):
                        setattr(entry, key, newname.rjust(5))
        else:
            # fix names
            oldname = self.get_correct_atomname(oldname)
            if len(newname) > 4:
                raise ValueError(f'{newname} is too long.')
            elif newname == 'END':
                self.log.info('I thing END may be an old keyword - What is ``ACT_COORD_ATOMS``?. BEST AVOID IT.')
            newname = self.pad_name(newname).upper()
            # find in atom.
            for atom in self.ATOM:
                if atom.name == newname:
                    raise ValueError(f'{newname} is already taken.')
                elif atom.name == oldname:
                    atom.name = newname
                    break
            # find in entries of the kind with ``first``, ``second``, ``third``, ``fourth`` attributes, which are atom names 4 char
            for attr in ('BOND', 'CHI', 'CUT_BOND'):  # ICOOR_INTERNAL ATOM_ALIAS
                for entry in getattr(self, attr):
                    for key in 'first', 'second', 'third', 'fourth':
                        if hasattr(entry, key) and getattr(entry, key) == oldname:
                            setattr(entry, key, newname)
                            break
            # find ICOOR_INTERNAL entries with ``child``,``parent``,``second_parent``,``third_parent`` attributes,
            # which are atom names 5 char
            for entry in self.ICOOR_INTERNAL:
                for key in 'child', 'parent', 'second_parent', 'third_parent':
                    if getattr(entry, key).strip() == oldname.strip():
                        setattr(entry, key, newname.ljust(5))
            for conn in self.CONNECT:
                if conn.atom_name.strip() == oldname.strip():
                    conn.atom_name = newname
            # find in the Generic entries
            for attr in ('ATOM_ALIAS', 'NBR_ATOM', 'FIRST_SIDECHAIN_ATOM', 'ADD_RING'):
                for entry in getattr(self, attr):
                    if oldname.strip() in entry.body:
                        entry.body = re.sub('(?<!\w)' + oldname.strip() + '(?!\w)', newname, entry.body)
            # find in the Generic list entries
            for attr in ('METAL_BINDING_ATOMS', 'ACT_COORD_ATOMS'):
                for entry in getattr(self, attr):
                    entry.values = [v if v.strip() != oldname.strip() else newname for v in entry.values]




########################################################################################################################
__doc__ = \
    """
The main class here is ``_ParamsIoMixin``, which adds load and dump to ``_ParamsInitMixin``.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################


import re
from .entries import Entries
from ._init_mixin import _ParamsInitMixin


class _ParamsIoMixin(_ParamsInitMixin):

    @classmethod
    def load(cls, filename: str):
        if hasattr(cls, '__name__'):
            self = cls()
        else:
            self = cls
        with open(filename, 'r') as w:
            for line in w:
                self._parse_line(line)
        return self

    @classmethod
    def loads(cls, text: str):
        if hasattr(cls, '__name__'):
            self = cls()
        else:
            self = cls
        for line in text.split('\n'):
            self._parse_line(line)
        return self

    def _parse_line(self, line):
        sline = line.strip()
        if not sline:
            return
        elif re.match('#', sline):
            header, body = re.match('(#+)\s?(.*)', sline).groups()
            self.comments.append(body)
        elif re.match('[A-Z]', sline):
            header, body = re.match('([_\w]+) (.*)', sline).groups()
            if '#' in body:
                body, comment = re.match('(.*?) ?#(.*)', body).groups()
                self.comments.append(comment)
            if 'CONNECT' in header:
                self.CONNECT.append(dict(atom_name=body,
                                         connect_type=header,
                                         index=len(self.CONNECT) + 1))
            elif header == 'NAME':
                self.NAME = body
            elif header in ('BOND', 'BOND_TYPE'):
                self.BOND.append(body)
            elif header in Entries.choices:
                getattr(self, header).append(body)
            else:
                raise ValueError(f'Not coded this far. {header}')
        else:
            self.comments.append(sline)

    def dumps(self) -> str:
        assert self.NBR_ATOM and self.NBR_ATOM[0], 'Undeclared NBR_ATOM entry'
        assert self.NBR_RADIUS and self.NBR_RADIUS[0], 'Undeclared NBR_RADIUS entry'
        lines = [f'NAME {self.NAME}', str(self.comments)]
        for entries in (self.IO_STRING, self.TYPE, self.AA, self.ROTAMER_AA, self.ATOM, self.ATOM_ALIAS, self.BOND,
                        self.CUT_BOND, self.ADD_RING, self.PROPERTIES, self.METAL_BINDING_ATOMS,
                        self.FIRST_SIDECHAIN_ATOM, self.RAMA_PREPRO_FILENAME, self.ACT_COORD_ATOMS, self.CHI,
                        self.CONNECT, self.NBR_ATOM, self.NBR_RADIUS, self.ICOOR_INTERNAL, self.PDB_ROTAMERS):
            if entries:
                lines.append(str(entries))
        return '\n'.join(lines)

    def dump(self, filename: str) -> None:
        with open(filename, 'w') as w:
            w.write(self.dumps())



from ._rdkit_inits import _RDKitInitMixin
from ._rdkit_rename import _RDKitRenameMixin
from ._rdkit_convert import _RDKitCovertMixin
from ._rdkit_prep import _RDKitPrepMixin

########################################################################################################################
__doc__ = \
    """
    The functionality is too big for a single class so was split into functional units.
    
    * **init** contains the various entry points
    * **rename** deals with atom renaming
    * **prep** labels atoms etc. for conversion
    * **convert** converts the mol into params entries
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

class _RDKitMixin(_RDKitInitMixin):
    # this is so that parts can change happily
    pass


from __future__ import annotations

from rdkit_to_params.entries import Entries

import os

########################################################################################################################
__doc__ = \
    """
The main class here is ``_RDKitPrepMixin``, which adds the various pre checks.
It does not rely on any ``Params`` entry stuff. So can be used by itself for testing.

    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

from typing import List, Dict, Union
from collections import defaultdict, deque, namedtuple
from rdkit import Chem
from rdkit.Chem import AllChem, rdFMCS
from warnings import warn
import re, string
from typing import Optional

from ._rdkit_rename import _RDKitRenameMixin

class _RDKitPrepMixin(_RDKitRenameMixin):

    def __init__(self):
        self.log.critical('This is init should not have been called! This exists to stop the IDE from getting angry/ for debugging.')
        self.NAME = 'LIG'
        self.TYPE = Entries.from_name('TYPE')
        self.mol = None
        self.generic = False
        self._rtype = []

    @classmethod
    def load_mol(cls, mol: Chem.Mol, generic:bool=False, name:Optional[str]=None) -> _RDKitPrepMixin:
        """
        A fully prepared molecule with optional dummy atoms to be coverted into a Params object

        :param mol: fully prepared molecule with optional dummy atoms
        :param generic: generic or classic atom types
        :param name: 3 letter name
        :return:
        """
        self = cls()
        self.log.debug('`load_mol` called...')
        self.mol = mol
        self.generic = generic
        self.TYPE.append('LIGAND')
        if name is not None:
            self.NAME = name
        self.fix_mol()
        # conversion elsewhere
        return self

    @classmethod
    def load_smiles(cls, smiles:str, generic:bool=False, name:Optional[str]=None) -> _RDKitPrepMixin:
        """
        A SMILES with optional dummy atoms to be coverted into a Params object

        :param smiles: can contain multiple *
        :param generic: generic or classic atom types
        :param name: 3 letter name
        :return:
        """
        cls.log.debug('`load_smiles` called...')
        mol = Chem.MolFromSmiles(smiles)
        mol.SetProp('_Name', name)
        mol = Chem.AddHs(mol)
        # cannot embed more than one dummy
        changed = []
        for atom in mol.GetAtoms():
            if atom.GetSymbol() == '*':
                atom.SetAtomicNum(6)
                atom.SetHybridization(Chem.HybridizationType.SP3)
                changed.append(atom.GetIdx())
        AllChem.EmbedMolecule(mol)
        AllChem.MMFFOptimizeMolecule(mol)
        AllChem.ComputeGasteigerCharges(mol)
        for atomIdx in changed:
            mol.GetAtomWithIdx(atomIdx).SetAtomicNum(0)
        # operate upon!
        return cls.load_mol(mol, generic, name)


    def fix_mol(self):
        self.log.debug('`fix_mol` called...')
        # partial charges.
        if not self.mol.GetAtomWithIdx(0).HasProp('_GasteigerCharge'):
            self.log.debug('... Adding Gasteiger')
            self._add_partial_charges()
        self._fix_atom_names()
        if self.generic is True:
            self._add_genrtypes()
        else:
            self._add_rtypes()

    def dump_pdb(self, filename: str, overwrite=False, stripped=True) -> None:
        """
        Write first conformer to a PDB file.

        :param filename: name of file.
        :param overwrite: error if exists?
        :param stripped: remove * atoms?
        :return: None
        """
        self.log.debug(f'Writing PDB to file {filename}')
        mol = self._prep_dump_pdb(filename, overwrite, stripped)
        Chem.MolToPDBFile(mol, filename)

    def dump_pdb_conf(self, filename: str, overwrite=False, stripped=True) -> int:
        """
        Write conformers to a PDB file.

        :param filename: name of file.
        :param overwrite: error if exists?
        :param stripped: remove * atoms?
        :return: number of conf written
        """
        self.log.debug(f'Writing PDB conformers to file {filename}')
        mol = self._prep_dump_pdb(filename, overwrite, stripped)
        w = Chem.PDBWriter(filename)
        for cid in range(mol.GetNumConformers()):
            w.write(mol, confId=cid)
        n = w.NumMols()
        w.close()
        return n

    def _prep_dump_pdb(self, filename: str, overwrite=False, stripped=True) -> Chem.Mol:
        if os.path.exists(filename) and overwrite:
            raise FileExistsError('The file already exists')
        if stripped:
            return self.dummyless
        else:
            return self.mol

    def dumps_pdb(self, stripped=True) -> str:
        if stripped:
            Chem.MolToPDBBlock(self.dummyless)
        else:
            Chem.MolToPDBBlock(self.mol)

    @property
    def dummyless(self):
        return Chem.DeleteSubstructs(self.mol, Chem.MolFromSmiles('*'))

    def _add_rtypes(self) -> None:
        """
        Add Rosetta Atom types to each atom.
        Mostly a guess...
        """

        # These are the elements with a single type.
        # Some have oxidation states '2p', but only one...
        element_to_type = {'B': 'Bsp2', 'F': 'F', 'CL': 'Cl', 'BR': 'Br', 'I': 'I', 'ZN': 'Zn2p', 'CO': 'Co2p',
                           'CU': 'Cu2p', 'MG': 'Mg2p', 'CA': 'Ca2p', 'Si': 'Si', 'NA': 'Na1p', 'K': 'K1p', 'HE': 'He',
                           'LI': 'Li', 'BE': 'Be', 'NE': 'Ne', 'AL': 'Al', 'AR': 'Ar', 'SC': 'Sc', 'TI': 'Ti', 'V': 'V',
                           'CR': 'Cr', 'MN': 'Mn', 'NI': 'Ni', 'GA': 'Ga', 'GE': 'Ge', 'AS': 'As', 'SE': 'Se',
                           'KR': 'Kr', 'RB': 'Rb', 'SR': 'Sr', 'Y': 'Y', 'ZR': 'Zr', 'NB': 'Nb', 'MO': 'Mo', 'TC': 'Tc',
                           'RU': 'Ru', 'RH': 'Rh', 'PD': 'Pd', 'AG': 'Ag', 'CD': 'Cd', 'IN': 'In', 'SN': 'Sn',
                           'SB': 'Sb', 'TE': 'Te', 'XE': 'Xe', 'CS': 'Cs', 'BA': 'Ba', 'LA': 'La', 'CE': 'Ce',
                           'PR': 'Pr', 'ND': 'Nd', 'PM': 'Pm', 'SM': 'Sm', 'EU': 'Eu', 'GD': 'Gd', 'TB': 'Tb',
                           'DY': 'Dy', 'HO': 'Ho', 'ER': 'Er', 'TM': 'Tm', 'YB': 'Yb', 'LU': 'Lu', 'HF': 'Hf',
                           'TA': 'Ta', 'W': 'W', 'RE': 'Re', 'OS': 'Os', 'IR': 'Ir', 'PT': 'Pt', 'AU': 'Au', 'HG': 'Hg',
                           'TL': 'Tl', 'PB': 'Pb', 'BI': 'Bi', 'PO': 'Po', 'AT': 'At', 'RN': 'Rn', 'FR': 'Fr',
                           'RA': 'Ra', 'AC': 'Ac', 'TH': 'Th', 'PA': 'Pa', 'U': 'U', 'NP': 'Np', 'PU': 'Pu', 'AM': 'Am',
                           'CM': 'Cm', 'BK': 'Bk', 'CF': 'Cf', 'ES': 'Es', 'FM': 'Fm', 'MD': 'Md', 'NO': 'No',
                           'LR': 'Lr'}

        groups = [
                  {'name': 'silane', 'SMARTS': '[Si]~O', 'types': ['Si', 'OSi']},
                  {'name': 'phosphoric', 'SMARTS': 'P[OH]', 'types': ['Pha', 'OHha', 'Hha']},
                  {'name': 'phosphate', 'SMARTS': 'P~O', 'types': ['Pha', 'OPha']},
                  {'name': 'free carbonic acid?', 'SMARTS': 'C(=O)(O)O', 'types': ['CO3', 'OC3', 'OC3']}, # no idea.
                  {'name': 'carboxylate', 'SMARTS': 'C(=O)[O-]', 'types': ['COO', 'OOC', 'OOC']},
                  {'name': 'carboxylic', 'SMARTS': 'C(=O)[OH]', 'types': ['COO', 'OOC', 'OH']},
                  {'name': 'ester', 'SMARTS': 'C(=O)OC', 'types': ['COO', 'Oet2', 'Oet3', None]},
                  {'name': 'ester?', 'SMARTS': 'C(=O)O', 'types': ['COO', 'OOC', 'OOC']},
                  {'name': '3amide', 'SMARTS': 'C(=O)N(C)C', 'types': ['CNH2', 'ONH2', 'Npro', None, None]},
                  {'name': '3amide', 'SMARTS': 'C(=O)NC', 'types': ['CNH2', 'ONH2', 'Nbb', None]},
                  {'name': 'amide', 'SMARTS': 'C(=O)N', 'types': ['CNH2', 'ONH2', 'NH2O']},
                  {'name': 'keto-aldehyde', 'SMARTS': 'C(=O)', 'types': [None, 'OOC']}, ## ???
                  {'name': '2amine', 'SMARTS': 'CNC', 'types': [None, 'Nbb', None]}, # actually Nbb is a amide.
                  {'name': 'azo', 'SMARTS': 'NN', 'types': ['NtrR', 'NtrR']},
                  {'name': 'azo', 'SMARTS': 'nn', 'types': ['NtrR', 'NtrR']},
                  {'name': 'aramid?', 'SMARTS': 'nc(o)c', 'types': ['NtrR', 'aroC', 'Oaro', None]}, # unsure how it's written
                  {'name': 'nitrile', 'SMARTS': 'C#N', 'types': ['COO', 'NtrR']}, ## the old one uses Nhis??!
                  {'name': 'imino', 'SMARTS': 'C=N', 'types': [None, 'Nhis']},
                  {'name': 'nitro', 'SMARTS': '[N+](=O)[O-]', 'types': ['Nhis','OOC', 'OOC']},
                  {'name': 'nitro_aro', 'SMARTS': 'n(=o)o', 'types': ['Nhis','OOC', 'OOC']},
                  {'name': 'furan', 'SMARTS': 'coc', 'types': ['aroC', 'Oaro', 'aroC']},
                  {'name': 'ether', 'SMARTS': 'COC', 'types': [None, 'Oet3', None]},
                  {'name': 'hydroxyl', 'SMARTS': '[OH]', 'types': ['OH']},
                  {'name': 'guanidium', 'SMARTS': 'NC(=N)N', 'types': ['Narg', 'aroC', 'Narg', 'Narg']},
                  ]
        for group in groups:
            template = Chem.MolFromSmarts(group['SMARTS'])
            template = Chem.AddHs(template, explicitOnly=True)
            types = group['types']
            for match in self.mol.GetSubstructMatches(template):
                for i, n in enumerate(types):
                    j = match[i]
                    atom = self.mol.GetAtomWithIdx(j)
                    if n is not None and not (atom.HasProp('_rType') and atom.GetProp('_rType').strip()):
                        atom.SetProp('_rType', n)
        # Simple atoms
        for atom in self.mol.GetAtoms():
            symbol = atom.GetSymbol()
            Hs = [a for a in atom.GetNeighbors() if a.GetSymbol() == 'H']
            if atom.HasProp('_rType') and atom.GetProp('_rType').strip():
                pass
            elif symbol == '*':
                atom.SetProp('_rType', 'VIRT')
            elif symbol == 'C':
                if atom.GetIsAromatic():
                    atom.SetProp('_rType', 'aroC')
                    for n in Hs:
                        n.SetProp('_rType', 'Haro')
                else:
                    atom.SetProp('_rType', 'CH' + str(len(Hs)))
            elif symbol == 'N':
                if atom.GetIsAromatic() and len(Hs) == 0:
                    atom.SetProp('_rType', 'Nhis')
                elif atom.GetIsAromatic():  # could also be NtrR...
                    atom.SetProp('_rType', 'Ntrp')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3 and len(Hs) < 3:
                    atom.SetProp('_rType', 'Nbb')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    atom.SetProp('_rType', 'Nlys')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2:
                    atom.SetProp('_rType', 'Npro')  # or Narg if the orbitals are funky...
                else:
                    raise ValueError(f'No idea what this nitrogen {atom.GetHybridization()} is')
            elif symbol == 'O':
                if atom.GetIsAromatic():
                    atom.SetProp('_rType', 'Oaro')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2:
                    atom.SetProp('_rType', 'Oet2')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    atom.SetProp('_rType', 'Oet3')
                else:
                    raise ValueError(f'No idea what this oxygen {atom.GetHybridization()} is')
            elif symbol == 'H':
                n = atom.GetNeighbors()[0]
                if n.GetSymbol() == 'C' and n.GetIsAromatic():
                    atom.SetProp('_rType', 'Haro')
                elif n.GetSymbol() == 'C':
                    atom.SetProp('_rType', 'Hapo')
                elif n.GetSymbol() == 'S':
                    atom.SetProp('_rType', 'HS')
                else:
                    atom.SetProp('_rType', 'Hpol')
            elif symbol == 'S':
                if len(Hs) == 1:
                    atom.SetProp('_rType', 'SH1')
                else:
                    atom.SetProp('_rType', 'S')
            elif symbol == 'P':
                atom.SetProp('_rType', 'Phos')
            elif symbol.upper() in element_to_type:
                atom.SetProp('_rType', element_to_type[symbol.upper()])
            else:
                self.log.warning(f'No idea what this {atom.GetSymbol()} {atom.GetHybridization()} is. assigning it REPLS')
                atom.SetProp('_rType', 'REPLS')

    def _add_genrtypes(self) -> None:
        """
        Add Rosetta Atom types to each atom.
        """
        groups = [{'name': 'carboxylate', 'SMARTS': 'C(=O)[O-]', 'types': ['COO', 'OOC', 'OOC']},
                  {'name': 'carboxylic', 'SMARTS': 'C(=O)[OH]', 'types': ['COO', 'OOC', 'OH']},
                  {'name': 'ester', 'SMARTS': 'C(=O)OC', 'types': ['COO', 'Oal', 'Oet', None]},
                  {'name': 'ester', 'SMARTS': 'C(=O)O', 'types': ['COO', 'OOC', 'OOC']},
                  {'name': '3amide', 'SMARTS': 'C(=O)[NH0]', 'types': ['CNH2', 'ONH2', 'Nad3']},
                  {'name': 'amide', 'SMARTS': 'C(=O)N', 'types': ['CNH2', 'ONH2', 'NH2O']},
                  {'name': 'keto-aldehyde', 'SMARTS': 'C(=O)', 'types': [None, 'Oal']},
                  {'name': '2amine', 'SMARTS': 'CNC', 'types': [None, 'Nam2', None]},
                  {'name': 'azo', 'SMARTS': 'NN', 'types': ['Nad3', 'Nad3']},
                  {'name': 'nitrile', 'SMARTS': 'C#N', 'types': ['CTp', 'NG1']},
                  {'name': 'imino', 'SMARTS': 'C=[NH]', 'types': [None, 'Nin']},
                  {'name': '2imino', 'SMARTS': 'C=[NH0]', 'types': [None, 'Nim']},
                  {'name': '2amine', 'SMARTS': 'CNC', 'types': [None, 'Nam', None]},
                  {'name': 'nitro', 'SMARTS': '[N+](=O)[O-]', 'types': ['NGb', 'Ont', 'Ont']},
                  {'name': 'nitro_aro', 'SMARTS': 'n(=o)o', 'types': ['NGb', 'Ont', 'Ont']},
                  {'name': 'furan', 'SMARTS': 'coc', 'types': ['aroC', 'Ofu', 'aroC']},
                  {'name': 'ether', 'SMARTS': 'COC', 'types': [None, 'Oet', None]},
                  {'name': 'hydroxyl', 'SMARTS': '[OH]', 'types': ['OH']},
                  {'name': 'guanidium', 'SMARTS': 'NC(=N)N', 'types': ['Narg', 'aroC', 'Narg', 'Narg']}
                  ]
        for group in groups:
            template = Chem.MolFromSmarts(group['SMARTS'])
            types = group['types']
            for match in self.mol.GetSubstructMatches(template):
                for i, n in enumerate(types):
                    j = match[i]
                    atom = self.mol.GetAtomWithIdx(j)
                    if n is not None and not (atom.HasProp('_rType') and atom.GetProp('_rType').strip()):
                        atom.SetProp('_rType', n)
        # Generic atoms
        for atom in self.mol.GetAtoms():
            symbol = atom.GetSymbol()
            Hs = [a for a in atom.GetNeighbors() if a.GetSymbol() == 'H']
            if atom.HasProp('_rType') and atom.GetProp('_rType').strip():
                pass
            elif symbol == '*':
                atom.SetProp('_rType', 'VIRT')
            elif symbol == 'C':
                if atom.GetIsAromatic():
                    atom.SetProp('_rType', 'aroC')
                    for n in Hs:
                        n.SetProp('_rType', 'Haro')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    if len(Hs) > 0:
                        atom.SetProp('_rType', 'CH' + str(len(Hs)))
                    else:
                        atom.SetProp('_rType', 'CS')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2:
                    if len(Hs) > 0:
                        atom.SetProp('_rType', 'CD' + str(len(Hs)))
                    else:
                        atom.SetProp('_rType', 'CD')
                elif atom.GetHybridization() == Chem.HybridizationType.SP:
                    atom.SetProp('_rType', 'CT')
                else:
                    raise ValueError(f'No idea what this carbon {atom.GetHybridization()} is')
            elif symbol == 'N':
                if atom.GetIsAromatic():
                    atom.SetProp('_rType', 'NGb')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3 and len(Hs) == 0:
                    atom.SetProp('_rType', 'NG3')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    atom.SetProp('_rType', 'Nam')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2 and len(Hs) == 0:
                    atom.SetProp('_rType', 'NG2')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2 and len(Hs) == 1:
                    atom.SetProp('_rType', 'NG21')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2 and len(Hs) == 2:
                    atom.SetProp('_rType', 'NG22')
                elif atom.GetHybridization() == Chem.HybridizationType.SP and len(Hs) == 0:
                    atom.SetProp('_rType', 'NG1')
                else:
                    raise ValueError(f'No idea what this nitrogen {atom.GetHybridization()} is')
            elif symbol == 'O':
                if atom.GetIsAromatic():
                    atom.SetProp('_rType', 'Oaro')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2 and len(Hs) == 0:
                    atom.SetProp('_rType', 'OG2')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3 and len(Hs) == 0:
                    atom.SetProp('_rType', 'OG3')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3 and len(Hs) == 1:
                    atom.SetProp('_rType', 'OG31')
                else:
                    raise ValueError(f'No idea what this oxygen {atom.GetHybridization()} is')
            elif symbol == 'H':
                n = atom.GetNeighbors()[0]
                if n.GetSymbol() == 'C' and n.GetIsAromatic():
                    atom.SetProp('_rType', 'Haro')
                elif n.GetSymbol() == 'C':
                    atom.SetProp('_rType', 'Hapo')
                elif n.GetSymbol() == 'S':
                    atom.SetProp('_rType', 'HS')
                elif n.GetSymbol() == 'N':
                    atom.SetProp('_rType', 'HN')
                elif n.GetSymbol() == 'O':
                    atom.SetProp('_rType', 'HO')
                else:
                    atom.SetProp('_rType', 'HG')
            elif symbol == 'S':
                if len(Hs) == 1:
                    atom.SetProp('_rType', 'Sth')
                elif len(Hs) == 0:
                    atom.SetProp('_rType', 'Ssl')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3 and atom.GetExplicitValence() == 6:
                    atom.SetProp('_rType', 'SG5')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    atom.SetProp('_rType', 'SG3')
                elif atom.GetHybridization() == Chem.HybridizationType.SP2:
                    atom.SetProp('_rType', 'SG2')
                else:
                    raise ValueError(f'No idea what this S {atom.GetHybridization()} is')
            elif symbol == 'P':
                if atom.GetHybridization() == Chem.HybridizationType.SP3 and atom.GetExplicitValence() == 6:
                    atom.SetProp('_rType', 'PG5')
                elif atom.GetHybridization() == Chem.HybridizationType.SP3:
                    atom.SetProp('_rType', 'PG3')
                else:
                    raise ValueError(f'No idea what this S {atom.GetHybridization()} is')
            elif symbol in ('F', 'Cl', 'Br', 'I'):
                n = atom.GetNeighbors()[0]
                if n.GetIsAromatic():
                    atom.SetProp('_rType', symbol + 'R')
                else:
                    atom.SetProp('_rType', symbol)
            else:
                raise ValueError(f'No idea what this {atom.GetSymbol()} {atom.GetHybridization()} is')

    def _aminoacid_override(self, elemental):
        aa = Chem.MolFromSmiles('*NCC(~O)*')
        if self.mol.HasSubstructMatch(aa):
            self.log.info('Ligand detected to be polymer!')
            self.TYPE.append('POLYMER')
            aa_map = dict([('LOWER', None),
                           (' N  ', 'Nbb'),
                           (' CA ', 'CAbb'),
                           (' C  ', 'CObb'),
                           (' O  ', 'OCbb'),
                           ('UPPER', None)
                           ])
            self.rename_by_substructure(aa, list(aa_map.keys()))
            # fix amine and H for secondary (proline)
            amine = self.get_atom_by_name(' N  ')
            hs = [n for n in amine.GetNeighbors() if n.GetSymbol() == 'H']
            if len(hs):
                self._set_PDBInfo_atomname(hs[0], ' H  ', overwrite=True)
                aa_map[' H  '] = 'HNbb'
            else:
                aa_map[' N  '] = 'Npro'
            ca = self.get_atom_by_name(' CA ')
            for neigh in ca.GetNeighbors():
                if neigh.GetSymbol() == 'H':
                    self.rename_atom(neigh, ' HA ')
            if self.greekification:
                self._greekify()
            # add rtypes
            self.retype_by_name(aa_map)
            # change conn
            elemental['CONN'] = 2  # when LOWER and UPPER exist the CONN is CONN3.

    def _greekify(self):
        greek = list('ABGDEZHTIKLMNXOPRS')
        greekdex = defaultdict(list)
        ca = self.get_atom_by_name('CA')
        for atom in self.mol.GetAtoms():
            is_backbone = (atom.GetPDBResidueInfo() is not None and
                    atom.GetPDBResidueInfo().GetName().strip() in ('LOWER', 'UPPER', 'N', 'CA', 'C', 'H', 'HA', 'O', 'OXT'))
            if atom.GetSymbol() != 'H' and not is_backbone:
                n = len(Chem.GetShortestPath(self.mol, ca.GetIdx(), atom.GetIdx())) - 1
                greekdex[n].append(atom)
        for k in greekdex:
            if len(greek) <= k:
                pass # finished the greek!
            elif len(greekdex[k]) == 0:
                pass # impossible tho
            elif len(greekdex[k]) == 1:
                name = f'{greekdex[k][0].GetSymbol(): >2}{greek[k]} '
                self.rename_atom(greekdex[k][0], name)
            elif len(greekdex[k]) < 36:
                l = list(range(1,10)) + list(string.ascii_uppercase)
                for i, atom in enumerate(greekdex[k]):
                    name = f'{atom.GetSymbol(): >2}{greek[k]}{l[i]}'
                    self.rename_atom(greekdex[k][0], name)
            else:
                pass # what????


    def _fix_atom_names(self):
        elemental = defaultdict(int)
        seen = []
        # Amino acid overwrite.
        self._aminoacid_override(elemental)
        for i in range(self.mol.GetNumAtoms()):
            atom = self.mol.GetAtomWithIdx(i)
            el = atom.GetSymbol().upper()
            if el == '*':
                el = 'CONN'
            elemental[el] += 1  # compatible mol_to_params.py
            lamename = el + str(elemental[el])
            lamename = self.pad_name(lamename, atom)
            while lamename in seen:
                elemental[el] += 1
                lamename = el + str(elemental[el])
            name = self._set_PDBInfo_atomname(atom, lamename, overwrite=False)
            if name in seen:
                self.log.warning(f'Name clash {name}, second one now called {lamename}')
                atom.GetPDBResidueInfo().SetName(lamename)
                seen.append(lamename)
            else:
                seen.append(name)

    def _add_partial_charges_OLD(self):
        """
        This is pointless convoluted.
        :return:
        """
        raise DeprecationWarning
        demovalence = {1: 1, 2: 8, 3: 7, 4: 6, 6: 15}
        # Not using Chem PeriodicTable because this was meant to be a quickfix...
        mol = Chem.Mol(self.mol)
        AllChem.ComputeGasteigerCharges(mol, throwOnParamFailure=False)
        tries = 0
        while True:
            if tries > 5:
                AllChem.ComputeGasteigerCharges(mol, throwOnParamFailure=False)
            try:
                AllChem.ComputeGasteigerCharges(mol, throwOnParamFailure=True)
            except ValueError as err:
                warn(f'{err.__class__.__name__}: {err}')
                dodgy = re.search('parameters for Element\: ([*\w]+)', str(err)).group(1)
                for atom in mol.GetAtoms():
                    if atom.GetSymbol() == dodgy:
                        valence = atom.GetExplicitValence()
                        atom.SetAtomicNum(demovalence[valence])
                tries += 1
            else:
                break
        for i in range(self.mol.GetNumAtoms()):
            gc = mol.GetAtomWithIdx(i).GetDoubleProp('_GasteigerCharge')
            if str(gc) != 'nan':
                self.mol.GetAtomWithIdx(i).SetDoubleProp('_GasteigerCharge', gc)
            else:
                self.mol.GetAtomWithIdx(i).SetDoubleProp('_GasteigerCharge', 0.)

    def _add_partial_charges(self):
        changed = []
        for atom in self.mol.GetAtoms():
            if atom.GetSymbol() == '*':
                atom.SetAtomicNum(6)
                atom.SetHybridization(Chem.HybridizationType.SP3)
                changed.append(atom.GetIdx())
        AllChem.ComputeGasteigerCharges(self.mol, throwOnParamFailure=False)
        for i, atom in enumerate(self.mol.GetAtoms()):
            if i in changed:
                atom.SetAtomicNum(0)
            gc = atom.GetDoubleProp('_GasteigerCharge')
            if str(gc) != 'nan':
                atom.SetDoubleProp('_GasteigerCharge', 0.)

    def _get_resn_from_PDBInfo(self):
        """
        Gets the residue name for PDB info.

        :return:
        """
        infos = [atom.GetPDBResidueInfo() for atom in self.mol.GetAtoms()]
        names = [i.GetResidueName() for i in infos if i is not None]
        if names:
            return names[0]
        else:
            return 'LIG'

    @classmethod
    def pad_name(self, name, atom=None):
        if name in ('CONN1','CONN2','CONN3','CONN4', 'LOWER', 'UPPER'):
            return name
        elif len(name) == 4:
            return name
        elif len(name) > 4:
            return name[:4]
        elif name[0] == ' ':
            return name.ljust(4)
        elif len(name) < 4 and (atom is None or len(atom.GetSymbol()) == 1):
            return ' ' + name.ljust(3)
        else:
            return name.ljust(4)

    def move_aside(self):
        """
        Changes the names of the atoms to not clash.

        :return:
        """
        for i in range(self.mol.GetNumAtoms()):
            self._set_PDBInfo_atomname(self.mol.GetAtomWithIdx(i), f'XX{i: <2}', overwrite=True)

    def add_Hs(self):
        """
        Add Hs before convert_mol step!
        :return:
        """
        self.log.debug('Adding hydrogens')
        self.mol = AllChem.AddHs(self.mol)
        changed = []
        for atom in self.mol.GetAtoms():
            if atom.GetSymbol() == '*':
                atom.SetAtomicNum(6)
                atom.SetHybridization(Chem.HybridizationType.SP3)
                changed.append(atom.GetIdx())
        AllChem.EmbedMolecule(self.mol, useRandomCoords=True)
        AllChem.MMFFOptimizeMolecule(self.mol)
        AllChem.ComputeGasteigerCharges(self.mol, throwOnParamFailure=False)
        for i, atom in enumerate(self.mol.GetAtoms()):
            if i in changed:
                atom.SetAtomicNum(0)
        self.fix_mol()



########################################################################################################################
__doc__ = \
    """
The main class here is `Entries``, which is a fancy list. It gets called for each uppercase attribute 
in the initialisation of ``Params`` (which happens in ``_ParamsInitMixin`` __e.g.__ ``Entries.from_name('IO_STRING')``).
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

from dataclasses import dataclass
import re
from warnings import warn

from collections import abc


########################################################################################################################

class Entries(abc.MutableSequence):
    """
    A fancy default list, where the elements are instances of whatver is in ``entry_cls``.
    It can be initialised via the class method ``from_name`` which accepst a string that has to be present in the class attribute ``.choices``.
    The ``.append`` method can work with str, list, dict or instance of the actual class it wants.
    Note that the string for the string way must be without the header to the line.
    The entry classes requires a ``from_str`` classmethod that returns an instance for this.
    They also require __str__ method as this is how the entries are converted into string.

    ``Entries.from_name('BOND')``
    """

    choices = {} # this gets filled after each class is declared.

    def __init__(self, entry_cls, singleton: bool = True):
        """
        The entries class is a fancy constrained list. The data is actually stored in ``.data``.

        :param entry_cls: what is the allowed class of the entries
        :param singleton: is only one entry allowed?
        """
        self.entry_cls = entry_cls
        self.singleton = singleton
        self.data = []

    @classmethod
    def from_name(cls, name: str):
        if name in cls.choices:
            cc, singleton = cls.choices[name]
            return cls(entry_cls=cc, singleton=singleton)

    def __getitem__(self, index):
        return self.data[index]

    def _assign_value(self, value):
        if isinstance(value, self.entry_cls):
            return value
        elif isinstance(value, str):
            return self.entry_cls.from_str(value)
        elif isinstance(value, list):
            return self.entry_cls(*value)
        elif isinstance(value, dict):
            return self.entry_cls(**value)
        else:
            raise ValueError(f'No idea what to do with {value}')

    def __setitem__(self, index, value):
        if self.singleton:
            index = 0
        self.data[index] = self._assign_value(value)

    def __delitem__(self, index):
        if self.singleton:
            index = 0
        del self.data[index]

    def insert(self, index, value):
        if self.singleton:
            self.data = [self._assign_value(value)]
        else:
            self.data.insert(index, self._assign_value(value))

    def __len__(self):
        return len(self.data)

    def __str__(self):
        lines = []
        for entry in self.data:
            lines.append(str(entry))
        return '\n'.join(lines)


#########################################################################################################

class GenericEntry:
    """
    This is meant to be inherited. ``header`` is the entry type. body is a string.
    """

    def __init__(self, header: str, body: str):
        self.header = header.strip().upper()
        self.body = body.rstrip()
        assert self.header, f'Type is empty'
        assert self.body, f'Value is empty'

    def __str__(self) -> str:
        return f'{self.header} {self.body}'

    @classmethod
    def from_str(cls, text):
        return cls(text)


#########################################################################################################

class GenericListEntry:
    """
    This is meant to be inherited. ``header`` is the entry type. ``values`` is a list of strings.
    """

    def __init__(self, header: str, *args: str):
        self.header = header.strip().upper()
        self.values = args

    def __str__(self) -> str:
        v = ' '.join(self.values)
        return f'{self.header} {v}'

    @classmethod
    def from_str(cls, text):
        return cls(*text.split())


#########################################################################################################

class NBR_ATOMEntry(GenericEntry):
    def __init__(self, body: str):
        super().__init__(header='NBR_ATOM', body=body)


Entries.choices['NBR_ATOM'] = (NBR_ATOMEntry, True)


#########################################################################################################

class NBR_RADIUSEntry(GenericEntry):
    def __init__(self, body: str):
        super().__init__(header='NBR_RADIUS', body=body)


Entries.choices['NBR_RADIUS'] = (NBR_RADIUSEntry, True)


#########################################################################################################

class CommentEntry(GenericEntry):
    def __init__(self, body: str):
        super().__init__(header='#', body=body)


Entries.choices['#'] = (CommentEntry, False)
Entries.choices['comment'] = (CommentEntry, False)


#########################################################################################################


class ATOM_ALIASEntry(GenericEntry):
    def __init__(self, body: str):
        super().__init__(header='ATOM_ALIAS', body=body)


Entries.choices['ATOM_ALIAS'] = (ATOM_ALIASEntry, False)


#########################################################################################################

@dataclass
class IO_STRINGEntry:
    """
    * ``.name3`` is three letter name. ``Params().NAME`` is actually a dynamic attribute that uses this.
    * ``.name1`` is a one letter name.

    These get checked for length.
    """
    name3: str = 'LIG'
    name1: str = 'Z'

    def __post_init__(self):
        assert len(self.name3) == 3, f'{self.name3} is not 3 char long'
        assert len(self.name1) == 1, f'{self.name1} is not 1 char long'

    def __str__(self) -> str:
        return f'IO_STRING {self.name3} {self.name1}'

    @classmethod
    def from_str(cls, text):
        name3, name1 = text.strip().split()
        return cls(name3, name1)


Entries.choices['IO_STRING'] = (IO_STRINGEntry, True)


#########################################################################################################

@dataclass
class CONNECTEntry:
    """
    This is a mess, but it guesses what you mean.
    Deals with UPPER, LOWER and CONNECT.

    """
    atom_name: str
    index: int = 1
    connect_type: str = ''  # | 'CONNECT' | 'UPPER_CONNECT' | 'LOWER_CONNECT'
    connect_name: str = ''  # 'CONN1' | 'UPPER' | 'LOWER'

    def __post_init__(self):
        if self.connect_type and self.connect_name:
            pass
        elif not self.connect_type and not self.connect_name:
            self.connect_type = 'CONNECT'
            self.connect_name = f'CONN{self.index}'
        elif not self.connect_type and 'CONN' not in self.connect_name:
            self.connect_type = f'{self.connect_name}_CONNECT'
        elif not self.connect_type:
            self.connect_type = 'CONNECT'
        elif not self.connect_name and self.connect_type == 'CONNECT':
            self.connect_name = f'CONN{self.index}'
        elif not self.connect_name:
            self.connect_name = self.connect_type.replace('_CONNECT', '')
        else:
            raise ValueError(f'I missed this case ({self.connect_name}, {self.connect_type}) in this badly written method')

    def __str__(self) -> str:
        return f'{self.connect_type} {self.atom_name}'

    @classmethod
    def from_str(cls, text):
        return cls(*text.split())


Entries.choices['CONNECT'] = (CONNECTEntry, False)


#########################################################################################################

@dataclass
class CHIEntry:
    index: int
    first: str
    second: str
    third: str
    fourth: str

    def __post_init__(self):
        self.fourth = self.fourth.ljust(4)

    def __str__(self) -> str:
        return f'CHI {self.index} {self.first} {self.second} {self.third} {self.fourth}'

    @classmethod
    def from_str(cls, text: str):
        # 1  C6   C5   C4   C3
        rex = re.match('(\d+) (.{4}) (.{4}) (.{4}) (.{2,4})', text)
        if rex is None:
            raise ValueError(f'CHI entry "{text}" is not formatted correctly')
        data = dict(zip(('index', 'first', 'second', 'third', 'fourth'), rex.groups()))
        return cls(**data)


Entries.choices['CHI'] = (CHIEntry, False)


#########################################################################################################


@dataclass
class ICOOR_INTERNALEntry:
    """
    Lines stolen from Rosetta documentation
    >                     Child  Phi Angle    Theta        Distance   Parent  Angle  Torsion
    >   ICOOR_INTERNAL    C14  167.536810   59.880644    1.473042   N2    C11   C12

    * Child atom (A4)
    * phi angle (torsion angle between A1, A2, A3, A4)
    * theta angle (improper angle = (180 - (angle between A4, A3, A2)))
    * distance (between A4 and A3)
    * parent atom (A3)
    * angle atom (A2)
    * torsion atom (A4)
    """
    child: str
    phi: float
    theta: float
    distance: float
    parent: str
    second_parent: str
    third_parent: str

    def __post_init__(self):
        self.third_parent = self.third_parent.ljust(5)

    def __str__(self) -> str:
        return f'ICOOR_INTERNAL  {self.child: <5} {self.phi: >11.6f} {self.theta: >11.6f} {self.distance: >11.6f} ' + \
               f'{self.parent: <5} {self.second_parent: <5} {self.third_parent: <5}'

    @classmethod
    def from_str(cls, text: str):
        rex = re.match(' (.{5}) (.{11}) (.{11}) (.{11}) (.{5}) (.{5}) (.{3,5})', text)
        if rex is None:
            raise ValueError(f'ICOOR_INTERNAL Entry "{text}" is not formatted correctly')
        data = list(rex.groups())
        for i in range(1, 4):
            data[i] = float(data[i])
        return cls(*data)


Entries.choices['ICOOR_INTERNAL'] = (ICOOR_INTERNALEntry, False)


#########################################################################################################


@dataclass
class BONDEntry:
    """
    dataclass class for both BOND and BOND_ENTRY. The ``__str__`` method will know based on ``.order``.
    The hash is the two atom names sorted. So BOND records with the same names will be equal.
    """
    first: str
    second: str
    order: int = 1  # 2,3, 4|ARO

    def __post_init__(self):
        self.second = self.second.ljust(4)

    def __str__(self) -> str:
        if self.order == 1 or not self.order:
            return f'BOND {self.first: >4} {self.second: >4}'
        else:
            return f'BOND_TYPE {self.first: >4} {self.second: >4} {self.order}'

    def __hash__(self):
        return hash('+'.join(sorted([self.first, self.second])))

    def __eq__(self, other):
        return hash(self) == hash(other)

    @classmethod
    def from_str(cls, text: str):
        rex = re.match('(.{4}) (.{2,4})\s?(.*)', text)
        if rex is None:
            raise ValueError(f'BOND entry "{text}" is not formatted correctly')
        data = dict(zip(('first', 'second', 'order'), rex.groups()))
        data['order'] = data['order'].strip()
        if data['order'] == '':
            data['order'] == 1
        elif data['order'] in ('ARO', '4'):
            data['order'] == 4  # ARO is also acceptable.
        elif isinstance(data['order'], int):
            pass
        else:
            data['order'] = int(data['order'].strip())
        return cls(**data)


Entries.choices['BOND'] = (BONDEntry, False)


#########################################################################################################


@dataclass
class ATOMEntry:
    # PDB atom name, Rosetta AtomType, MM AtomType, and charge
    name: str
    rtype: str
    mtype: str = 'X'
    partial: float = 0

    def __str__(self) -> str:
        return f'ATOM {self.name: >4} {self.rtype: >4} {self.mtype: >4} {self.partial:.7f}'

    def __eq__(self, other):
        """
        ``atomentry == 'CA'`` will return false because ``atomentry.name.strip() == 'CA'`` will return true.
        """
        if isinstance(other, self.__class__):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)


    @classmethod
    def from_str(cls, text: str):
        rex = re.match('(.{4}) (.{4}) (.{4}) +([-\d\.]+)', text)
        if rex is None:
            raise ValueError(f'ATOM entry "{text}" is not formatted correctly')
        data = dict(zip(('name', 'rtype', 'mtype', 'partial'), rex.groups()))
        data['partial'] = float(data['partial'])
        return cls(**data)


Entries.choices['ATOM'] = (ATOMEntry, False)


#########################################################################################################

@dataclass
class CUT_BONDEntry:
    """
    No idea what CUT_BOND is for.
    """
    first: str
    second: str

    def __post_init__(self):
        self.second = self.second.ljust(4)

    def __str__(self) -> str:
        return f'CUT_BOND {self.first: >4} {self.second: >4}'

    @classmethod
    def from_str(cls, text: str):
        rex = re.match('(.{4}) (.{2,4})', text)
        if rex is None:
            raise ValueError(f'CUT_BOND entry "{text}" is not formatted correctly')
        data = dict(zip(('first', 'second'), rex.groups()))
        return cls(**data)

Entries.choices['CUT_BOND'] = (CUT_BONDEntry, False)


#########################################################################################################

class PDB_ROTAMERSEntry(GenericEntry):
    """
    This does zero checks for fine existance.
    """
    def __init__(self, body: str):
        super().__init__(header='PDB_ROTAMERS', body=body)


Entries.choices['PDB_ROTAMERS'] = (PDB_ROTAMERSEntry, True)


#########################################################################################################

class ROTAMER_AAEntry(GenericEntry):
    def __init__(self, body: str):
        super().__init__(header='ROTAMER_AA', body=body)


Entries.choices['ROTAMER_AA'] = (ROTAMER_AAEntry, True)


#########################################################################################################

class AAEntry(GenericEntry):
    def __init__(self, body: str = 'UNK'):
        if body != 'UNK':
            self.log.info('AA should be UNK... tolerating oddity.')
        super().__init__(header='AA', body=body)


Entries.choices['AA'] = (AAEntry, True)


#########################################################################################################

class TYPEEntry(GenericEntry):
    """
    LIGAND or POLYMER. No exceptions.
    """
    def __init__(self, body: str = 'LIGAND'):
        assert body in ('POLYMER', 'LIGAND'), f'residue TYPE {body} is neither POLYMER or LIGAND'
        super().__init__(header='TYPE', body=body)


Entries.choices['TYPE'] = (TYPEEntry, True)


#########################################################################################################

class ADD_RINGEntry(GenericEntry):
    ## To be fixed. Spacing drama...
    def __init__(self, body: str):
        self.log.info('ADD_RING is sloppily coded. the values are stored as an unsplit string!')
        super().__init__(header='ADD_RING', body=body)


Entries.choices['ADD_RING'] = (ADD_RINGEntry, False)


#########################################################################################################

class PROPERTIESEntry(GenericListEntry):
    def __init__(self, *args: str):
        super().__init__('PROPERTIES', *args)


Entries.choices['PROPERTIES'] = (PROPERTIESEntry, False)


#########################################################################################################

class FIRST_SIDECHAIN_ATOMEntry(GenericEntry):
    def __init__(self, body:str):
        super().__init__(header='FIRST_SIDECHAIN_ATOM', body=body)


Entries.choices['FIRST_SIDECHAIN_ATOM'] = (FIRST_SIDECHAIN_ATOMEntry, True)


#########################################################################################################

class RAMA_PREPRO_FILENAMEEntry(GenericEntry):
    def __init__(self, body:str):
        super().__init__(header='RAMA_PREPRO_FILENAME', body=body)


Entries.choices['RAMA_PREPRO_FILENAME'] = (RAMA_PREPRO_FILENAMEEntry, True)


#########################################################################################################

class METAL_BINDING_ATOMSEntry(GenericListEntry):
    def __init__(self, *args:str):
        super().__init__('METAL_BINDING_ATOMS', *args)


Entries.choices['METAL_BINDING_ATOMS'] = (METAL_BINDING_ATOMSEntry, True)


#########################################################################################################

class ACT_COORD_ATOMSEntry(GenericListEntry):
    def __init__(self, *args: str):
        super().__init__('ACT_COORD_ATOMS', *args)


Entries.choices['ACT_COORD_ATOMS'] = (ACT_COORD_ATOMSEntry, True)


#########################################################################################################


########################################################################################################################

__doc__ = \
"""
``_PoserMixin`` is a base that adds pyrosetta functionality if avaliable. Adds:

*the bound method ``.test(outfile=None)``, which returns a pose.
*the static method ``.params_to_pose(paramsfile:str, name:str)``, which returns a pose.

"""
__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################


import pyrosetta, os

from typing import Optional

class _PoserMixin:
    def test(self, outfile:Optional[str]=None) -> pyrosetta.Pose:
        """
        Makes a pose with the ligand to see if it works.

        :param outfile: optionally save file.
        :return: pose
        """
        self.log.debug(f'Testing params {outfile}')
        pose = self.to_pose(relax=True)
        assert pose.total_residue() == 1, 'Residue failed.'
        if outfile:
            pose.dump_pdb(outfile)
        return pose

    def to_pose(self, relax=False):
        pose = pyrosetta.Pose()
        # add paramsblock
        rts = self.add_residuetype(pose)
        # add new residue
        lig = pyrosetta.rosetta.core.conformation.ResidueFactory.create_residue(rts.name_map('LIG'))
        pose.append_residue_by_jump(lig, 1)
        if relax:
            cycles = 15
            scorefxn = pyrosetta.get_fa_scorefxn()
            relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, cycles)
            relax.apply(pose)
        return pose

    def add_residuetype(self, pose: pyrosetta.Pose) -> pyrosetta.rosetta.core.chemical.ResidueTypeSet:
        rts = pose.conformation().modifiable_residue_type_set_for_conf(pyrosetta.rosetta.core.chemical.FULL_ATOM_t)
        buffer = pyrosetta.rosetta.std.stringbuf(self.dumps())
        stream = pyrosetta.rosetta.std.istream(buffer)
        new = pyrosetta.rosetta.core.chemical.read_topology_file(stream, self.NAME,
                                                                 rts)  # no idea what the second argument does
        rts.add_base_residue_type(new)
        return rts


    @staticmethod
    def params_to_pose(paramsfile:str, name:str) -> pyrosetta.Pose:
        """
        Staticmethod to get a pose from a params file.

        :param paramsfile: params file.
        :param name: 3-letter residue name.
        :return:
        """
        pose = pyrosetta.Pose()
        params_paths = pyrosetta.rosetta.utility.vector1_string()
        params_paths.extend([paramsfile])
        resiset = pyrosetta.generate_nonstandard_residue_set(pose, params_paths)
        lig = pyrosetta.rosetta.core.conformation.ResidueFactory.create_residue( rts.name_map( name ) )
        pose.append_residue_by_jump(lig, 1)
        return pose



########################################################################################################################
__doc__ = \
    """
    Pip install file. All the files have this block here, so in PyCharm I can change them en-mass.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.1"
__citation__ = "None."

########################################################################################################################


import setuptools
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

print('rdkit and pyrosetta are optional... but are the main reason this was written.')
print('The former is conda, the latter is downloaded from RosettaCommons.')

setuptools.setup(
                name='rdkit_to_params',
                version=__version__,
                python_requires='>3.6',
                packages=setuptools.find_packages(),
                url='https://github.com/matteoferla/rdkit_to_params',
                license='MIT',
                author=__author__,
                author_email=__email__,
                description='Create or modify Rosetta params files (topology files) from scratch, RDKit mols or another params file.',
                long_description=long_description,
                long_description_content_type='text/markdown',
                test_suite='tests',
                install_requires=[
                ],
                extras_require = {
                        'rdkit_feature':  ["rdkit"],
                        'testing_feature': ["pyrosetta"]
                    }
)



from __future__ import annotations

########################################################################################################################
__doc__ = \
    """
    THis the core conversion. the method ``convert_mol`` does something you'd never have guessed. 
    There are probably loads of mistakes here ---or more correctly
    in ``_RDKitPrepMixin()._add_rtypes`` or ``_RDKitPrepMixin()._add_genrtypes``.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

from ._rdkit_prep import _RDKitPrepMixin

from typing import List, Dict, Union, Optional
from collections import defaultdict, deque, namedtuple
from rdkit import Chem
from rdkit.Chem import AllChem
from warnings import warn
import random

import numpy as np


class _RDKitCovertMixin(_RDKitPrepMixin):
    _Measure = namedtuple('Measure', ['distance', 'angle', 'torsion'])

    greekification = True  # controls whether to change the atom names to greek for amino acids for CB and beyond.

    def convert_mol(self):
        """
        This method does the actual conversion to params entries

        :return:
        """
        # NAME
        if self.mol.HasProp("_Name"):
            title = self.mol.GetProp("_Name").strip()
        else:
            title = 'Unnamed molecule'
        if len(title) == 3:
            self.NAME = title
        else:
            self.comments.append(title)
            self.NAME = self._get_resn_from_PDBInfo()
        self.log.debug(f'{title} is being converted (`.convert_mol`)')
        assert len(Chem.GetMolFrags(self.mol)) == 1, f'{title} is split in {len(Chem.GetMolFrags(self.mol))}'
        # SMILES
        self.comments.append(Chem.MolToSmiles(self.mol))
        if len(self.TYPE) == 0:
            self.TYPE.append('LIGAND')
        # remove all previous entries.
        self.ATOM.data = []
        self.CUT_BOND.data = []
        self.CHI.data = []
        self.BOND.data = []
        # ICOOR
        self._parse_icoors()  # fills self.ordered_atoms
        # ATOM & CONNECT
        self._parse_atoms()
        # CHI
        self._parse_rotatables()
        # BOND
        self._parse_bonds()

        # NBR
        self._find_centroid()

    ## ============= internal coordinates ==============================================================================

    def _parse_icoors(self) -> None:
        self.log.debug(f'Filling ICOOR')
        self._undescribed = deque(self.mol.GetAtoms())
        self.ordered_atoms = []
        self._rotation_count = 0
        if self.is_aminoacid():
            # you have to start with N.
            N = self.get_atom_by_name('N')
            CA = self.get_atom_by_name('CA')
            C = self.get_atom_by_name('C')
            O = self.get_atom_by_name('O')
            UPPER = self.get_atom_by_name('UPPER')
            LOWER = self.get_atom_by_name('LOWER')
            self._add_icoor([N, N, CA, C])
            self._add_icoor([CA, N, CA, C])
            self._add_icoor([C, CA, N, C])
            self._add_icoor([UPPER, C, CA, N])
            self._add_icoor([O, C, CA, UPPER])
            self._add_icoor([LOWER, N, CA, C])
        else:
            # cycle through until a good root is found!
            while self._undescribed:
                self._prevent_overcycling()
                atom = self._undescribed[0]
                if atom.GetSymbol() == '*':
                    self.log.debug(f'Dummy not allowed in first three lines...')
                    self._undescribed.rotate(-1)
                    self._rotation_count += 1
                    continue  # This cannot be in the first three lines, whereas the R group is first in a canonical SMILES.
                else:
                    # I assume in some cases Virtual atoms are called for... but for now, no!
                    for parent in self._get_unseen_neighbors(atom, [], True):
                        siblings = self._get_unseen_neighbors(parent, [atom], True)
                        if len(siblings) != 0:
                            break
                    else:
                        self.log.debug(f'First row: No siblings for {self._get_PDBInfo_atomname(atom)}!')
                        self._undescribed.rotate(-1)
                        self._rotation_count += 1
                        continue
                    grandparent = siblings[0]
                    # first line
                    self._add_icoor([atom, atom, parent, grandparent])
                    # Second line
                    self._add_icoor([parent, atom, parent, grandparent])
                    # Third line
                    self._add_icoor([grandparent, parent, atom, grandparent])
                    break
        self.log.debug('Parsing forth line onwards')
        # not first three lines
        while self._undescribed:
            self._prevent_overcycling()
            atom = self._undescribed[0]
            d_idx = [d.GetIdx() for d in self.ordered_atoms]
            d_neighs = [n for n in self._get_unseen_neighbors(atom, [], False) if n.GetIdx() in d_idx]
            if len(d_neighs) == 0:
                self.log.debug(f'Other row: No parent for {self._get_PDBInfo_atomname(atom)}!')
                self._undescribed.rotate(-1)
                self._rotation_count += 1
                continue
            parent = d_neighs[0]
            d_neighs = [n for n in self._get_unseen_neighbors(parent, [atom], False) if n.GetIdx() in d_idx]
            if len(d_neighs) == 0:
                self._undescribed.rotate(-1)
                self.log.debug(f'Other row: No siblings for  {self._get_PDBInfo_atomname(atom)}!')
                self._rotation_count += 1
                continue
            elif len(d_neighs) == 1:
                sibling = d_neighs[0]
                d_neighs = [n for n in self._get_unseen_neighbors(sibling, [parent, atom], False) if
                            n.GetIdx() in d_idx]
                if len(d_neighs) == 0:
                    self._undescribed.rotate(-1)
                    self.log.debug(f'Other row: No cousins for  {self._get_PDBInfo_atomname(atom)}')
                    self._rotation_count += 1
                    continue
                else:
                    atoms = [atom, parent, sibling, d_neighs[0]]
            else:
                atoms = [atom, parent, d_neighs[0], d_neighs[1]]
            self._add_icoor(atoms)
            self._rotation_count -= 1

    def _add_icoor(self, atoms):
        # due to madness the same atom objects differ.
        self._undescribed.remove([this for this in self._undescribed if atoms[0].GetIdx() == this.GetIdx()][0])
        self.ordered_atoms.append(atoms[0])
        m = self._get_measurements(self.mol.GetConformer(), *atoms)
        self.ICOOR_INTERNAL.append(dict(child=self._get_PDBInfo_atomname(atoms[0]),
                                        phi=m.torsion,
                                        theta=m.angle,
                                        distance=m.distance,
                                        parent=self._get_PDBInfo_atomname(atoms[1]),
                                        second_parent=self._get_PDBInfo_atomname(atoms[2]),
                                        third_parent=self._get_PDBInfo_atomname(atoms[3])))

    def _prevent_overcycling(self):
        # prevent overcycling...
        self.log.debug(f'Current deque rotation count {self._rotation_count}')
        if self._rotation_count > 5 * self.mol.GetNumAtoms():
            d = [f'{a.GetIdx()}: {self._get_PDBInfo_atomname(a)}' for a in self.ordered_atoms]
            u = [f'{a.GetIdx()}: {self._get_PDBInfo_atomname(a)}' for a in self._undescribed]
            raise StopIteration(f'Too many cycles. described: {d}, undescribed: {u}')
        elif self._rotation_count % self.mol.GetNumAtoms() == 0:
            # it has done a full rotation.
            if self._rotation_count != 0:
                self.log.debug(f'Ordering atoms with a shuffle first')
                random.shuffle(self._undescribed)  # mix up equal tier ones
            else:
                self.log.debug(f'Ordering atoms without a shuffle first')
            if len(self.ordered_atoms) != 0:
                self.log.debug(f'Ordering atoms by distance from root, and being not hydrogen or dummy')
                root = self.ordered_atoms[0]
                distance_penaltifier = lambda atom: len(Chem.GetShortestPath(self.mol, root.GetIdx(), atom.GetIdx()))
            else:
                self.log.debug(f'Ordering atoms by being not hydrogen or dummy')
                distance_penaltifier = lambda atom: 0
            hydrogen_penaltifier = lambda atom: 0 if atom.GetSymbol() != 'H' else 100
            dummy_penaltifier = lambda atom: 0 if atom.GetSymbol() != '*' else 200
            scorer = lambda atom: distance_penaltifier(atom) + \
                                  hydrogen_penaltifier(atom) + \
                                  dummy_penaltifier(atom)
            self._undescribed = deque(sorted(self._undescribed, key=scorer))

    def _get_measurements(self, conf: Chem.Conformer, a: Chem.Atom, b: Chem.Atom, c: Chem.Atom, d: Chem.Atom):
        ai = a.GetIdx()
        bi = b.GetIdx()
        ci = c.GetIdx()
        di = d.GetIdx()
        dist = 0
        angle = 0
        tor = 0
        try:
            dist = Chem.rdMolTransforms.GetBondLength(conf, ai, bi)
            angle = 180 - Chem.rdMolTransforms.GetAngleDeg(conf, ai, bi, ci)
            tor = Chem.rdMolTransforms.GetDihedralDeg(conf, ai, bi, ci, di)
        except ValueError:
            pass
        if str(tor) == 'nan':  # quicker than isnan.
            tor = 0
        return self._Measure(distance=dist,
                             angle=angle,
                             torsion=tor)

    ## ============= Atom entries ======================================================================================

    def _parse_atoms(self):
        self.log.debug(f'Filling ATOM')
        for atom in self.ordered_atoms:
            self._parse_atom(atom)

    def _parse_atom(self, atom: Chem.Atom) -> None:
        if atom.GetSymbol() == '*':
            neighbor = atom.GetNeighbors()[0]
            n_name = self._get_PDBInfo_atomname(neighbor)
            if self.is_aminoacid() and neighbor.GetSymbol() == 'N':
                # atom_name, index, connect_type, connect_name
                self.CONNECT.append([n_name, 1, 'LOWER_CONNECT', 'LOWER'])
            elif self.is_aminoacid() and neighbor.GetSymbol() == 'C':
                # atom_name, index, connect_type, connect_name
                self.CONNECT.append([n_name, 2, 'UPPER_CONNECT', 'UPPER'])
            elif self.is_aminoacid():
                i = max(3, len(self.CONNECT) + 1)
                self.CONNECT.append([n_name, i, 'CONNECT'])
            else:
                self.CONNECT.append([n_name, len(self.CONNECT) + 1, 'CONNECT'])
        else:
            d = self._get_atom_descriptors(atom)
            self.ATOM.append(d)

    ## ============= Chi entries =======================================================================================

    def _parse_rotatables(self):
        """
        I am not sure if this is right...
        I do not know if a C(=O)-C=O is rotatable (cis-trans).

        :return:
        """
        self.log.debug(f'Filling CHI')

        def is_single(a, b):
            bond = self.mol.GetBondBetweenAtoms(a.GetIdx(), b.GetIdx())
            return bond.GetBondType() == Chem.BondType.SINGLE

        cuts = [tuple(sorted([ring_set[0], ring_set[-1]])) for ring_set in self.mol.GetRingInfo().AtomRings()]

        def is_cut(a, b):
            pair = tuple(sorted([a.GetIdx(), b.GetIdx()]))
            return pair in cuts

        ordered_indices = [a.GetIdx() for a in self.ordered_atoms]
        get_place = lambda atom: ordered_indices.index(atom.GetIdx())

        for atom in self.ordered_atoms:
            if atom.GetSymbol() == 'H':
                continue  # PROTON_CHI 3 SAMPLES 2 0 180 EXTRA 1 20 thing...
            elif atom.GetSymbol() == '*':
                continue
            for neighbor in self._get_unseen_neighbors(atom, []):
                if get_place(atom) > get_place(neighbor):
                    continue
                if is_cut(atom, neighbor):
                    continue  # dont cross a cut!
                elif is_single(atom, neighbor):
                    for grandneighbor in self._get_unseen_neighbors(neighbor, [atom]):
                        if is_cut(neighbor, grandneighbor):
                            continue
                        if is_single(grandneighbor, neighbor):
                            ggneighbors = self._get_unseen_neighbors(grandneighbor, [atom, neighbor])
                            if ggneighbors:  # 3-4 bond does not need to be single, nor does it matter which is the 4th atom
                                self.CHI.append(dict(index=len(self.CHI) + 1,
                                                     first=self._get_PDBInfo_atomname(atom),
                                                     second=self._get_PDBInfo_atomname(neighbor),
                                                     third=self._get_PDBInfo_atomname(grandneighbor),
                                                     fourth=self._get_PDBInfo_atomname(ggneighbors[0])))

    def _get_unseen_neighbors(self, atom: Chem.Atom, seen: List[Chem.Atom], nondummy: bool = True):
        neighbors = atom.GetNeighbors()
        if nondummy:
            neighbors = [neighbor for neighbor in neighbors if neighbor.GetSymbol() != '*']
        seenIdx = {a.GetIdx() for a in seen}
        return [neighbor for neighbor in neighbors if neighbor.GetIdx() not in seenIdx]

    ## ============= Bond entries ======================================================================================

    def _parse_bonds(self):
        self.log.debug(f'Filling BOND')
        for bond in self.mol.GetBonds():
            self._parse_bond(bond)
        for ring_set in self.mol.GetRingInfo().AtomRings():
            self.log.debug(f'Mol has a ring, adding a `CUT_BOND` entry (`.convert_mol`)')
            self.CUT_BOND.append({'first': self._get_PDBInfo_atomname(self.mol.GetAtomWithIdx(ring_set[0])),
                                  'second': self._get_PDBInfo_atomname(self.mol.GetAtomWithIdx(ring_set[-1]))})

    def _parse_bond(self, bond: Chem.Bond) -> None:
        if any([atom.GetSymbol() == '*' for atom in (bond.GetBeginAtom(), bond.GetEndAtom())]):
            return None  # CONNECT.
        if bond.GetBondTypeAsDouble() == 1.5:
            order = 4
        else:
            order = int(bond.GetBondTypeAsDouble())
        begin = bond.GetBeginAtom()
        end = bond.GetEndAtom()
        self.BOND.append([self._get_PDBInfo_atomname(begin),
                          self._get_PDBInfo_atomname(end),
                          order])

    def _get_atom_descriptors(self, atom: Chem.Atom) -> dict:
        return {'name': self._get_PDBInfo_atomname(atom),
                'rtype': atom.GetProp('_rType'),
                'mtype': ' X  ',
                'partial': atom.GetDoubleProp('_GasteigerCharge')}

    def _get_nondummy_neighbors(self, atom) -> List[str]:
        # returns list of names!
        return [self._get_PDBInfo_atomname(neighbor) for neighbor in atom.GetNeighbors() if neighbor.GetSymbol() != '*']

    def _find_centroid(self):
        a = Chem.Get3DDistanceMatrix(self.mol)
        n = np.linalg.norm(a, axis=0)  # Frobenius norm
        i = int(np.argmin(n))
        s = np.max(a, axis=0)[i]
        self.NBR_ATOM.append(self.mol.GetAtomWithIdx(i).GetPDBResidueInfo().GetName())
        self.NBR_RADIUS.append(str(s))

    def polish_mol(self, resi=1, chain='X'):
        """
        The mol may be inconsistent in its PDBResidueInfo
        :return:
        """
        resn = self.NAME[:3]
        self.log.debug(f'Making all PDBResidueInfo resn={resn} resi={resi} chain={chain}')
        for atom in self.mol.GetAtoms():
            info = atom.GetPDBResidueInfo()
            info.SetResidueName(resn)
            info.SetResidueNumber(resi)
            info.SetChainId(chain)
            #info.SetSegmentNumber(segi) commented out because it is an int not string
            if self.is_aminoacid:
                info.SetIsHeteroAtom(False)
            else:
                info.SetIsHeteroAtom(True)



from __future__ import annotations
########################################################################################################################
__doc__ = \
    """
    Various entry points.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.0"
__citation__ = "None."

########################################################################################################################

from rdkit import Chem
from rdkit.Chem import AllChem
from typing import Optional, Dict, List
from ._rdkit_convert import _RDKitCovertMixin

import warnings

class _RDKitInitMixin(_RDKitCovertMixin):

    @classmethod
    def from_mol(cls, mol: Chem.Mol,
                 name: Optional[str] = None,
                 generic: bool = False,
                 atomnames: Optional[Dict[int, str]] = None) -> _RDKitInitMixin:
        """
        :param mol: Rdkit molecule, with explicit protons and all.
        :type mol: Chem.Mol
        :param name: 3-letter name
        :type name: str
        :param generic: convert mol with generic or classic AtomTypes?
        :type generic: bool
        :param atomnames: optional dictionary to set names.
        :rtype: instance
        """
        cls.log.debug('`from_mol` called...')
        self = cls.load_mol(mol, generic, name)  # stores and calls .fix_mol() method in prep file.
        self.convert_mol()
        return self

    @classmethod
    def from_smiles(cls, smiles: str, name='LIG', generic: bool = False,
                    atomnames: Optional[Dict[int, str]] = None) -> _RDKitInitMixin:
        """
        Make a Params instance from a smiles.

        :param smiles: SMILES to use.
        :param name: name3/resn
        :param generic: use generic atoms types
        :param atomnames: optional dictionary to set names.
        :return:
        """
        cls.log.debug('`from_smiles` called...')
        self = cls.load_smiles(smiles=smiles, name=name, generic=generic)
        if atomnames is None:
            pass
        else:
            self.rename(atomnames)
            self.polish_mol()
        self.convert_mol()
        return self

    @classmethod
    def from_smiles_w_pdbblock(cls, pdb_block: str, smiles: str, generic: bool = False, name='LIG',
                              proximityBonding: bool = True):
        """
        Assumes there is only one residue of the resn/name3

        :param pdb_block:
        :param generic: convert mol with generic or classic AtomTypes?
        :type generic: bool
        :param name: 3-letter name
        :type name: str
        :param proximityBonding: rdkit option for Chem.MolFromPDBFile. Did the author of the pdb **not** add CONECT?
        :type proximityBonding: bool
        :rtype: instance
        """
        cls.log.debug('`from_smiles_w_pdbblock` called...')
        pdb = Chem.MolFromPDBBlock(pdb_block, removeHs=False, proximityBonding=proximityBonding)
        return cls._from_smiles_w_pdb(pdb, smiles, generic, name)


    @classmethod
    def from_smiles_w_pdbfile(cls, pdb_file: str, smiles: str, generic: bool = False, name='LIG',
                              proximityBonding: bool = True):
        """
        Assumes there is only one residue of the resn/name3

        :param pdb_file:
        :param generic: convert mol with generic or classic AtomTypes?
        :type generic: bool
        :param name: 3-letter name
        :type name: str
        :param proximityBonding: rdkit option for Chem.MolFromPDBFile. Did the author of the pdb **not** add CONECT?
        :type proximityBonding: bool
        :rtype: instance
        """
        cls.log.debug('`from_smiles_w_pdbfile` called...')
        pdb = Chem.MolFromPDBFile(pdb_file, removeHs=False, proximityBonding=proximityBonding)
        return cls._from_smiles_w_pdb(pdb, smiles, generic, name)

    @classmethod
    def _from_smiles_w_pdb(cls, pdb: Chem.Mol, smiles, generic, name):
        dodgy = Chem.SplitMolByPDBResidues(pdb, whiteList=[name])[name]
        AllChem.SanitizeMol(dodgy)
        good = Chem.MolFromSmiles(smiles)
        good.SetProp('_Name', name)
        dummies = []
        for atom in good.GetAtoms():
            if atom.GetSymbol() == '*':
                atom.SetAtomicNum(9)
                dummies.append(atom.GetIdx())
        good = AllChem.AddHs(good)
        AllChem.EmbedMolecule(good)
        AllChem.ComputeGasteigerCharges(good)
        AllChem.MMFFOptimizeMolecule(good)
        for d in dummies:
            good.GetAtomWithIdx(d).SetAtomicNum(0)
        self = cls.load_mol(good, generic=generic, name=name)
        self.move_aside()
        self.rename_from_template(dodgy)
        self.convert_mol()
        #####
        warnings.warn('CHI DISABLED. - has issues with this mode')  # todo correct this issue!
        self.CHI = [] # !!!!
        return self